@@ -1,4 +1,4 @@
-/* $Id: display.c,v 1.25 2002/10/29 16:19:41 ukai Exp $ */
+/* $Id: display.c,v 1.64 2003/02/06 17:16:56 ukai Exp $ */
 #include <signal.h>
 #include "fm.h"
 
@@ -164,7 +164,7 @@ fmTerm(void)
 	refresh();
 #ifdef USE_IMAGE
 	if (activeImage)
-	    loadImage(IMG_FLAG_STOP);
+	    loadImage(NULL, IMG_FLAG_STOP);
 #endif
 #ifdef USE_MOUSE
 	if (use_mouse)
@@ -214,30 +214,175 @@ static Linecolor color_mode = 0;
 static Buffer *save_current_buf = NULL;
 #endif
 
-char *delayed_msg = NULL;
+static char *delayed_msg = NULL;
 
+static void drawAnchorCursor(Buffer *buf);
+#define redrawBuffer(buf) redrawNLine(buf, LASTLINE)
+static void redrawNLine(Buffer *buf, int n);
+static Line *redrawLine(Buffer *buf, Line *l, int i);
 #ifdef USE_IMAGE
 static int image_touch = 0;
 static int draw_image_flag = FALSE;
 static Line *redrawLineImage(Buffer *buf, Line *l, int i);
 #endif
+static int redrawLineRegion(Buffer *buf, Line *l, int i, int bpos, int epos);
+static void do_effects(Lineprop m);
+#ifdef USE_ANSI_COLOR
+static void do_color(Linecolor c);
+#endif
+
+static Str
+make_lastline_link(Buffer *buf, char *title, char *url)
+{
+    Str s = NULL, u;
+#ifdef JP_CHARSET
+    Lineprop *pr;
+#endif
+    ParsedURL pu;
+    char *p;
+    int l = COLS - 1, i;
+
+    if (title && *title) {
+	s = Strnew_m_charp("[", title, "]", NULL);
+	for (p = s->ptr; *p; p++) {
+	    if (IS_CNTRL(*p) || IS_SPACE(*p))
+		*p = ' ';
+	}
+	if (url)
+	    Strcat_charp(s, " ");
+	l -= s->length;
+	if (l <= 0)
+	    return s;
+    }
+    if (!url)
+	return s;
+    parseURL2(url, &pu, baseURL(buf));
+    u = parsedURL2Str(&pu);
+    if (DecodeURL)
+	u = Strnew_charp(url_unquote_conv(u->ptr, buf->document_code));
+#ifdef JP_CHARSET
+    u = checkType(u, &pr, NULL);
+#endif
+    if (l <= 4 || l >= u->length) {
+	if (!s)
+	    return u;
+	Strcat(s, u);
+	return s;
+    }
+    if (!s)
+	s = Strnew_size(COLS);
+    i = (l - 2) / 2;
+#ifdef JP_CHARSET
+    if (CharType(pr[i]) == PC_KANJI2)
+	i--;
+#endif
+    Strcat_charp_n(s, u->ptr, i);
+#if LANG == JA
+    Strcat_charp(s, "");
+#else				/* LANG != JA */
+    Strcat_charp(s, "..");
+#endif				/* LANG != JA */
+    i = u->length - (COLS - 1 - s->length);
+#ifdef JP_CHARSET
+    if (CharType(pr[i]) == PC_KANJI2)
+	i++;
+#endif
+    Strcat_charp(s, &u->ptr[i]);
+    return s;
+}
+
+static Str
+make_lastline_message(Buffer *buf)
+{
+    Str msg, s = NULL;
+
+    if (displayLink) {
+#ifdef USE_IMAGE
+	MapArea *a = retrieveCurrentMapArea(buf);
+	if (a)
+	    s = make_lastline_link(buf, a->alt, a->url);
+	else
+#endif
+	{
+	    Anchor *a = retrieveCurrentAnchor(buf);
+	    char *p = NULL;
+	    if (a && a->title && *a->title)
+		p = a->title;
+	    else {
+		Anchor *a_img = retrieveCurrentImg(buf);
+		if (a_img && a_img->title && *a_img->title)
+		    p = a_img->title;
+	    }
+	    if (p || a)
+		s = make_lastline_link(buf, p, a ? a->url : NULL);
+	}
+	if (s && s->length >= COLS - 3)
+	    return s;
+    }
+
+#ifdef USE_MOUSE
+    if (use_mouse && mouse_action.lastline_str)
+	msg = Strnew_charp(mouse_action.lastline_str);
+    else
+#endif				/* not USE_MOUSE */
+	msg = Strnew();
+    if (displayLineInfo && buf->currentLine != NULL && buf->lastLine != NULL) {
+	int cl = buf->currentLine->real_linenumber;
+	int ll = buf->lastLine->real_linenumber;
+	int r = (int)((double)cl * 100.0 / (double)(ll ? ll : 1) + 0.5);
+	Strcat(msg, Sprintf("%d/%d (%d%%)", cl, ll, r));
+    }
+    else
+	Strcat_charp(msg, "Viewing");
+#ifdef USE_SSL
+    if (buf->ssl_certificate)
+	Strcat_charp(msg, "[SSL]");
+#endif
+    Strcat_charp(msg, " <");
+    Strcat_charp(msg, buf->buffername);
+
+    if (s) {
+	int l = COLS - 3 - s->length;
+	if (msg->length > l) {
+#ifdef JP_CHARSET
+	    char *p;
+	    int i;
+	    for (p = msg->ptr; *p; p += i) {
+		i = get_mclen(get_mctype(p));
+		l -= i;
+		if (l < 0)
+		    break;
+	    }
+	    l = p - msg->ptr;
+#endif
+	    Strtruncate(msg, l);
+	}
+	Strcat_charp(msg, "> ");
+	Strcat(msg, s);
+    }
+    else {
+	Strcat_charp(msg, ">");
+    }
+    return msg;
+}
 
 void
 displayBuffer(Buffer *buf, int mode)
 {
     Str msg;
-    Anchor *aa = NULL;
+    int ny = 0;
 
     if (buf->topLine == NULL && readBufferCache(buf) == 0) {	/* clear_buffer */
 	mode = B_FORCE_REDRAW;
     }
 
     if (buf->width == 0)
-	buf->width = COLS;
+	buf->width = INIT_BUFFER_WIDTH;
     if (buf->height == 0)
 	buf->height = LASTLINE + 1;
-    if ((buf->width != INIT_BUFFER_WIDTH && buf->type &&
-	 !strcmp(buf->type, "text/html")) || buf->need_reshape) {
+    if ((buf->width != INIT_BUFFER_WIDTH &&
+	 ((buf->type && !strcmp(buf->type, "text/html")) || FoldLine))
+	|| buf->need_reshape) {
 	buf->need_reshape = TRUE;
 	reshapeBuffer(buf);
     }
@@ -253,24 +398,39 @@ displayBuffer(Buffer *buf, int mode)
     else
 	buf->rootX = 0;
     buf->COLS = COLS - buf->rootX;
-    if (mode == B_FORCE_REDRAW || mode == B_SCROLL ||
-#ifdef USE_IMAGE
-	mode == B_REDRAW_IMAGE ||
+    if (nTab > 1
+#ifdef USE_MOUSE
+	|| mouse_action.menu_str
 #endif
+	) {
+	if (mode == B_FORCE_REDRAW || mode == B_REDRAW_IMAGE)
+	    calcTabPos();
+	ny = LastTab->y + 2;
+	if (ny > LASTLINE)
+	    ny = LASTLINE;
+    }
+    if (buf->rootY != ny || buf->LINES != LASTLINE - ny) {
+	buf->rootY = ny;
+	buf->LINES = LASTLINE - ny;
+	arrangeCursor(buf);
+	mode = B_REDRAW_IMAGE;
+    }
+    if (mode == B_FORCE_REDRAW || mode == B_SCROLL || mode == B_REDRAW_IMAGE ||
 	cline != buf->topLine || ccolumn != buf->currentColumn) {
+#ifdef USE_RAW_SCROLL
 	if (
 #ifdef USE_IMAGE
 	       !(activeImage && displayImage && draw_image_flag) &&
 #endif
 	       mode == B_SCROLL && cline && buf->currentColumn == ccolumn) {
 	    int n = buf->topLine->linenumber - cline->linenumber;
-	    if (n > 0 && n < LASTLINE) {
+	    if (n > 0 && n < buf->LINES) {
 		move(LASTLINE, 0);
 		clrtoeolx();
 		refresh();
 		scroll(n);
 	    }
-	    else if (n < 0 && n > -LASTLINE) {
+	    else if (n < 0 && n > -buf->LINES) {
 #if defined(__CYGWIN__) && LANG == JA
 		move(LASTLINE + n + 1, 0);
 		clrtoeolx();
@@ -280,7 +440,9 @@ displayBuffer(Buffer *buf, int mode)
 	    }
 	    redrawNLine(buf, n);
 	}
-	else {
+	else
+#endif
+	{
 #ifdef USE_IMAGE
 	    if (activeImage &&
 		(mode == B_REDRAW_IMAGE ||
@@ -288,7 +450,7 @@ displayBuffer(Buffer *buf, int mode)
 		if (draw_image_flag)
 		    clear();
 		clearImage();
-		loadImage(IMG_FLAG_STOP);
+		loadImage(buf, IMG_FLAG_STOP);
 		image_touch++;
 		draw_image_flag = FALSE;
 	    }
@@ -308,79 +470,11 @@ displayBuffer(Buffer *buf, int mode)
     }
 #endif
 
-#ifdef USE_MOUSE
-    if (use_mouse)
-#if LANG == JA
-	msg = Strnew_charp("㢬");
-#else				/* LANG != JA */
-	msg = Strnew_charp("<=UpDn ");
-#endif				/* LANG != JA */
-    else
-#endif				/* not USE_MOUSE */
-	msg = Strnew();
-    if (displayLineInfo && buf->currentLine != NULL && buf->lastLine != NULL) {
-	int cl = buf->currentLine->real_linenumber;
-	int ll = buf->lastLine->real_linenumber;
-	int r = (int)((double)cl * 100.0 / (double)ll + 0.5);
-	Strcat(msg, Sprintf("%d/%d (%d%%)", cl, ll, r));
-    }
-    else
-	Strcat_charp(msg, "Viewing");
-#ifdef USE_SSL
-    if (buf->ssl_certificate)
-	Strcat_charp(msg, "[SSL]");
-#endif
-    Strcat_charp(msg, " <");
-    Strcat_charp(msg, buf->buffername);
-    if (displayLink)
-	aa = retrieveCurrentAnchor(buf);
-    if (aa) {
-	ParsedURL url;
-	Str s;
-	int l;
-	parseURL2(aa->url, &url, baseURL(buf));
-	s = parsedURL2Str(&url);
-	l = buf->width - 2;
-	if (s->length > l) {
-	    if (l >= 4) {
-		msg = Strsubstr(s, 0, (l - 2) / 2);
-#if LANG == JA
-		Strcat_charp(msg, "");
-#else				/* LANG != JA */
-		Strcat_charp(msg, "..");
-#endif				/* LANG != JA */
-		l = buf->width - msg->length;
-		Strcat(msg, Strsubstr(s, s->length - l, l));
-	    }
-	    else {
-		msg = s;
-	    }
-	}
-	else {
-	    l -= s->length;
-	    if (msg->length > l) {
-#ifdef JP_CHARSET
-		char *bn = msg->ptr;
-		int i, j;
-		for (i = 0; bn[i]; i += j) {
-		    j = get_mclen(get_mctype(&bn[i]));
-		    if (i + j > l)
-			break;
-		}
-		l = i;
-#endif
-		Strtruncate(msg, l);
-	    }
-	    Strcat_charp(msg, "> ");
-	    Strcat(msg, s);
-	}
-    }
-    else {
-	Strcat_charp(msg, ">");
-    }
+    drawAnchorCursor(buf);
+
+    msg = make_lastline_message(buf);
     if (buf->firstLine == NULL) {
 	Strcat_charp(msg, "\tNo Line");
-	clear();
     }
     if (delayed_msg != NULL) {
 	disp_message(delayed_msg, FALSE);
@@ -388,15 +482,12 @@ displayBuffer(Buffer *buf, int mode)
 	refresh();
     }
     standout();
-    message(msg->ptr, buf->cursorX + buf->rootX, buf->cursorY);
+    message(msg->ptr, buf->cursorX + buf->rootX, buf->cursorY + buf->rootY);
     standend();
-    term_title(buf->buffername);
+    term_title(conv_to_system(buf->buffername));
     refresh();
 #ifdef USE_IMAGE
     if (activeImage && displayImage && buf->img) {
-	/*
-	 * loadImage(IMG_FLAG_START);
-	 */
 	drawImage();
     }
 #endif
@@ -408,16 +499,86 @@ displayBuffer(Buffer *buf, int mode)
 #endif
 }
 
-void
-redrawBuffer(Buffer *buf)
+static void
+drawAnchorCursor0(Buffer *buf, AnchorList *al, int hseq, int prevhseq,
+		  int tline, int eline, int active)
 {
-    redrawNLine(buf, LASTLINE);
+    int i, j;
+    Line *l;
+    Anchor *an;
+
+    l = buf->topLine;
+    for (j = 0; j < al->nanchor; j++) {
+	an = &al->anchors[j];
+	if (an->start.line < tline)
+	    continue;
+	if (an->start.line >= eline)
+	    return;
+	for (;; l = l->next) {
+	    if (l == NULL)
+		return;
+	    if (l->linenumber == an->start.line)
+		break;
+	}
+	if (hseq >= 0 && an->hseq == hseq) {
+	    for (i = an->start.pos; i < an->end.pos; i++) {
+		if (l->propBuf[i] & (PE_IMAGE | PE_ANCHOR | PE_FORM)) {
+		    if (active)
+			l->propBuf[i] |= PE_ACTIVE;
+		    else
+			l->propBuf[i] &= ~PE_ACTIVE;
+		}
+	    }
+	    if (active)
+		redrawLineRegion(buf, l, l->linenumber - tline + buf->rootY,
+				 an->start.pos, an->end.pos);
+	}
+	else if (prevhseq >= 0 && an->hseq == prevhseq) {
+	    if (active)
+		redrawLineRegion(buf, l, l->linenumber - tline + buf->rootY,
+				 an->start.pos, an->end.pos);
+	}
+    }
 }
 
-void
+static void
+drawAnchorCursor(Buffer *buf)
+{
+    Anchor *an;
+    int hseq, prevhseq;
+    int tline, eline;
+
+    if (!buf->firstLine || !buf->hmarklist)
+	return;
+    if (!buf->href && !buf->formitem)
+	return;
+
+    an = retrieveCurrentAnchor(buf);
+    if (!an)
+	an = retrieveCurrentMap(buf);
+    if (an)
+	hseq = an->hseq;
+    else
+	hseq = -1;
+    tline = buf->topLine->linenumber;
+    eline = tline + buf->LINES;
+    prevhseq = buf->hmarklist->prevhseq;
+
+    if (buf->href) {
+	drawAnchorCursor0(buf, buf->href, hseq, prevhseq, tline, eline, 1);
+	drawAnchorCursor0(buf, buf->href, hseq, -1, tline, eline, 0);
+    }
+    if (buf->formitem) {
+	drawAnchorCursor0(buf, buf->formitem, hseq, prevhseq, tline, eline, 1);
+	drawAnchorCursor0(buf, buf->formitem, hseq, -1, tline, eline, 0);
+    }
+    buf->hmarklist->prevhseq = hseq;
+}
+
+static void
 redrawNLine(Buffer *buf, int n)
 {
-    Line *l, *l0;
+    Line *l;
     int i;
 
 #ifdef USE_COLOR
@@ -428,32 +589,68 @@ redrawNLine(Buffer *buf, int n)
 #endif				/* USE_BG_COLOR */
     }
 #endif				/* USE_COLOR */
-    for (i = 0, l = buf->topLine; i < LASTLINE; i++) {
-	if (i >= LASTLINE - n || i < -n)
-	    l0 = redrawLine(buf, l, i);
-	else {
-	    l0 = (l) ? l->next : NULL;
+    if (nTab > 1
+#ifdef USE_MOUSE
+	|| mouse_action.menu_str
+#endif
+	) {
+	TabBuffer *t;
+	int l;
+
+	move(0, 0);
+#ifdef USE_MOUSE
+	if (mouse_action.menu_str)
+	    addstr(mouse_action.menu_str);
+#endif
+	clrtoeolx();
+	for (t = FirstTab; t; t = t->nextTab) {
+	    move(t->y, t->x1);
+	    if (t == CurrentTab)
+		bold();
+	    addch('[');
+	    l = t->x2 - t->x1 - 1 - strlen(t->currentBuffer->buffername);
+	    if (l < 0)
+		l = 0;
+	    if (l / 2 > 0)
+		addnstr_sup(" ", l / 2);
+	    if (t == CurrentTab)
+		EFFECT_ACTIVE_START;
+	    addnstr(t->currentBuffer->buffername, t->x2 - t->x1 - l);
+	    if (t == CurrentTab)
+		EFFECT_ACTIVE_END;
+	    if ((l + 1) / 2 > 0)
+		addnstr_sup(" ", (l + 1) / 2);
+	    move(t->y, t->x2);
+	    addch(']');
+	    if (t == CurrentTab)
+		boldend();
 	}
-	if (l0 == NULL && l == NULL)
-	    break;
-	l = l0;
+#if 0
+	move(0, COLS - 2);
+	addstr(" x");
+#endif
+	move(LastTab->y + 1, 0);
+	for (i = 0; i < COLS; i++)
+	    addch('~');
     }
-    if (n > 0)
+    for (i = 0, l = buf->topLine; i < buf->LINES; i++, l = l->next) {
+	if (i >= buf->LINES - n || i < -n)
+	    l = redrawLine(buf, l, i + buf->rootY);
+	if (l == NULL)
+	    break;
+    }
+    if (n > 0) {
+	move(i + buf->rootY, 0);
 	clrtobotx();
+    }
 
 #ifdef USE_IMAGE
     if (!(activeImage && displayImage && buf->img))
 	return;
-    move(buf->cursorY, buf->cursorX);
-    for (i = 0, l = buf->topLine; i < LASTLINE; i++) {
-	if (i >= LASTLINE - n || i < -n)
-	    l0 = redrawLineImage(buf, l, i);
-	else {
-	    l0 = (l) ? l->next : NULL;
-	}
-	if (l0 == NULL && l == NULL)
-	    break;
-	l = l0;
+    move(buf->cursorY + buf->rootY, buf->cursorX + buf->rootX);
+    for (i = 0, l = buf->topLine; i < buf->LINES && l; i++, l = l->next) {
+	if (i >= buf->LINES - n || i < -n)
+	    redrawLineImage(buf, l, i + buf->rootY);
     }
     getAllImage(buf);
 #endif
@@ -461,7 +658,7 @@ redrawNLine(Buffer *buf, int n)
 
 #define addKanji(pc,pr) (addChar((pc)[0],(pr)[0]),addChar((pc)[1],(pr)[1]))
 
-Line *
+static Line *
 redrawLine(Buffer *buf, Line *l, int i)
 {
     int j, pos, rcol, ncol, delta;
@@ -479,7 +676,7 @@ redrawLine(Buffer *buf, Line *l, int i)
 
     if (l == NULL) {
 	if (buf->pagerSource) {
-	    l = getNextPage(buf, LASTLINE - i);
+	    l = getNextPage(buf, buf->LINES + buf->rootY - i);
 	    if (l == NULL)
 		return NULL;
 	}
@@ -499,7 +696,7 @@ redrawLine(Buffer *buf, Line *l, int i)
 		buf->rootX = COLS;
 	    buf->COLS = COLS - buf->rootX;
 	}
-	if (l->real_linenumber)
+	if (l->real_linenumber && !l->bpos)
 	    sprintf(tmp, "%*ld:", buf->rootX - 1, l->real_linenumber);
 	else
 	    sprintf(tmp, "%*s ", buf->rootX - 1, "");
@@ -510,7 +707,7 @@ redrawLine(Buffer *buf, Line *l, int i)
 	l->width = COLPOS(l, l->len);
     if (l->len == 0 || l->width - 1 < column) {
 	clrtoeolx();
-	return l->next;
+	return l;
     }
     /* need_clrtoeol(); */
     pos = columnPos(l, column);
@@ -626,11 +823,11 @@ redrawLine(Buffer *buf, Line *l, int i)
 #endif
     if (rcol - column < buf->COLS)
 	clrtoeolx();
-    return l->next;
+    return l;
 }
 
 #ifdef USE_IMAGE
-Line *
+static Line *
 redrawLineImage(Buffer *buf, Line *l, int i)
 {
     int j, pos, rcol;
@@ -643,7 +840,7 @@ redrawLineImage(Buffer *buf, Line *l, int i)
     if (l->width < 0)
 	l->width = COLPOS(l, l->len);
     if (l->len == 0 || l->width - 1 < column)
-	return l->next;
+	return l;
     pos = columnPos(l, column);
     rcol = COLPOS(l, pos);
     for (j = 0; rcol - column < buf->COLS && pos + j < l->len; j++) {
@@ -696,11 +893,11 @@ redrawLineImage(Buffer *buf, Line *l, int i)
 	}
 	rcol = COLPOS(l, pos + j + 1);
     }
-    return l->next;
+    return l;
 }
 #endif
 
-int
+static int
 redrawLineRegion(Buffer *buf, Line *l, int i, int bpos, int epos)
 {
     int j, pos, rcol, ncol, delta;
@@ -853,7 +1050,7 @@ if (modeflag) { \
     modeflag = FALSE; \
 }
 
-void
+static void
 do_effects(Lineprop m)
 {
     /* effect end */
@@ -898,7 +1095,7 @@ do_effects(Lineprop m)
 }
 
 #ifdef USE_ANSI_COLOR
-void
+static void
 do_color(Linecolor c)
 {
     if (c & 0x8)
@@ -951,7 +1148,7 @@ addChar(char c, Lineprop mode)
 	addch(' ');
 }
 
-GeneralList *message_list = NULL;
+static GeneralList *message_list = NULL;
 
 void
 record_err_message(char *s)
@@ -998,6 +1195,13 @@ message(char *s, int return_x, int return_y)
     move(return_y, return_x);
 }
 
+void
+disp_err_message(char *s, int redraw_current)
+{
+    record_err_message(s);
+    disp_message(s, redraw_current);
+}
+
 void
 disp_message_nsec(char *s, int redraw_current, int sec, int purge, int mouse)
 {
@@ -1007,9 +1211,9 @@ disp_message_nsec(char *s, int redraw_current, int sec, int purge, int mouse)
 	fprintf(stderr, "%s\n", conv_to_system(s));
 	return;
     }
-    if (Currentbuf != NULL)
+    if (CurrentTab != NULL && Currentbuf != NULL)
 	message(s, Currentbuf->cursorX + Currentbuf->rootX,
-		Currentbuf->cursorY);
+		Currentbuf->cursorY + Currentbuf->rootY);
     else
 	message(s, LASTLINE, 0);
     refresh();
@@ -1022,7 +1226,7 @@ disp_message_nsec(char *s, int redraw_current, int sec, int purge, int mouse)
     if (mouse && use_mouse)
 	mouse_inactive();
 #endif
-    if (Currentbuf != NULL && redraw_current)
+    if (CurrentTab != NULL && Currentbuf != NULL && redraw_current)
 	displayBuffer(Currentbuf, B_NORMAL);
 }
 
@@ -1046,10 +1250,8 @@ set_delayed_message(char *s)
 }
 
 void
-cursorUp(Buffer *buf, int n)
+cursorUp0(Buffer *buf, int n)
 {
-    if (buf->firstLine == NULL)
-	return;
     if (buf->cursorY > 0)
 	cursorUpDown(buf, -1);
     else {
@@ -1061,11 +1263,28 @@ cursorUp(Buffer *buf, int n)
 }
 
 void
-cursorDown(Buffer *buf, int n)
+cursorUp(Buffer *buf, int n)
 {
+    Line *l = buf->currentLine;
     if (buf->firstLine == NULL)
 	return;
-    if (buf->cursorY < LASTLINE - 1)
+    while (buf->currentLine->prev && buf->currentLine->bpos)
+	cursorUp0(buf, n);
+    if (buf->currentLine == buf->firstLine) {
+	gotoLine(buf, l->linenumber);
+	arrangeLine(buf);
+	return;
+    }
+    cursorUp0(buf, n);
+    while (buf->currentLine->prev && buf->currentLine->bpos &&
+	   buf->currentLine->bwidth >= buf->currentColumn + buf->visualpos)
+	cursorUp0(buf, n);
+}
+
+void
+cursorDown0(Buffer *buf, int n)
+{
+    if (buf->cursorY < buf->LINES - 1)
 	cursorUpDown(buf, 1);
     else {
 	buf->topLine = lineSkip(buf, buf->topLine, n, FALSE);
@@ -1075,6 +1294,26 @@ cursorDown(Buffer *buf, int n)
     }
 }
 
+void
+cursorDown(Buffer *buf, int n)
+{
+    Line *l = buf->currentLine;
+    if (buf->firstLine == NULL)
+	return;
+    while (buf->currentLine->next && buf->currentLine->next->bpos)
+	cursorDown0(buf, n);
+    if (buf->currentLine == buf->lastLine) {
+	gotoLine(buf, l->linenumber);
+	arrangeLine(buf);
+	return;
+    }
+    cursorDown0(buf, n);
+    while (buf->currentLine->next && buf->currentLine->next->bpos &&
+	   buf->currentLine->bwidth + buf->currentLine->width <
+	   buf->currentColumn + buf->visualpos)
+	cursorDown0(buf, n);
+}
+
 void
 cursorUpDown(Buffer *buf, int n)
 {
@@ -1096,7 +1335,7 @@ cursorRight(Buffer *buf, int n)
 
     if (buf->firstLine == NULL)
 	return;
-    if (buf->pos == l->len)
+    if (buf->pos == l->len && !(l->next && l->next->bpos))
 	return;
     i = buf->pos;
     p = l->propBuf;
@@ -1110,6 +1349,12 @@ cursorRight(Buffer *buf, int n)
     else if (l->len == 0) {
 	buf->pos = 0;
     }
+    else if (l->next && l->next->bpos) {
+	cursorDown0(buf, 1);
+	buf->pos = 0;
+	arrangeCursor(buf);
+	return;
+    }
     else {
 	buf->pos = l->len - 1;
 #ifdef JP_CHARSET
@@ -1118,7 +1363,7 @@ cursorRight(Buffer *buf, int n)
 #endif				/* JP_CHARSET */
     }
     cpos = COLPOS(l, buf->pos);
-    buf->visualpos = cpos - buf->currentColumn;
+    buf->visualpos = l->bwidth + cpos - buf->currentColumn;
     delta = 1;
 #ifdef JP_CHARSET
     if (CharType(p[buf->pos]) == PC_KANJI1)
@@ -1127,9 +1372,9 @@ cursorRight(Buffer *buf, int n)
     vpos2 = COLPOS(l, buf->pos + delta) - buf->currentColumn - 1;
     if (vpos2 >= buf->COLS && n) {
 	columnSkip(buf, n + (vpos2 - buf->COLS) - (vpos2 - buf->COLS) % n);
-	buf->visualpos = cpos - buf->currentColumn;
+	buf->visualpos = l->bwidth + cpos - buf->currentColumn;
     }
-    buf->cursorX = buf->visualpos;
+    buf->cursorX = buf->visualpos - l->bwidth;
 }
 
 void
@@ -1147,17 +1392,25 @@ cursorLeft(Buffer *buf, int n)
     if (i >= 2 && CharType(p[i - 1]) == PC_KANJI2)
 	delta = 2;
 #endif				/* JP_CHARSET */
-    if (i > delta)
+    if (i >= delta)
 	buf->pos = i - delta;
+    else if (l->prev && l->bpos) {
+	cursorUp0(buf, -1);
+	buf->pos = buf->currentLine->len - 1;
+	arrangeCursor(buf);
+	return;
+    }
     else
 	buf->pos = 0;
     cpos = COLPOS(l, buf->pos);
-    buf->visualpos = cpos - buf->currentColumn;
-    if (buf->visualpos < 0 && n) {
-	columnSkip(buf, -n + buf->visualpos - buf->visualpos % n);
-	buf->visualpos = cpos - buf->currentColumn;
+    buf->visualpos = l->bwidth + cpos - buf->currentColumn;
+    if (buf->visualpos - l->bwidth < 0 && n) {
+	columnSkip(buf,
+		   -n + buf->visualpos - l->bwidth - (buf->visualpos -
+						      l->bwidth) % n);
+	buf->visualpos = l->bwidth + cpos - buf->currentColumn;
     }
-    buf->cursorX = buf->visualpos;
+    buf->cursorX = buf->visualpos - l->bwidth;
 }
 
 void
@@ -1175,17 +1428,31 @@ cursorHome(Buffer *buf)
 void
 arrangeCursor(Buffer *buf)
 {
-    int col, col2;
+    int col, col2, pos;
     int delta = 1;
     if (buf == NULL || buf->currentLine == NULL)
 	return;
     /* Arrange line */
-    if (buf->currentLine->linenumber - buf->topLine->linenumber >= LASTLINE ||
-	buf->currentLine->linenumber < buf->topLine->linenumber) {
-	buf->topLine = buf->currentLine;
+    if (buf->currentLine->linenumber - buf->topLine->linenumber >= buf->LINES
+	|| buf->currentLine->linenumber < buf->topLine->linenumber) {
+	/*
+	 * buf->topLine = buf->currentLine;
+	 */
+	buf->topLine = lineSkip(buf, buf->currentLine, 0, FALSE);
     }
     /* Arrange column */
-    if (buf->currentLine->len == 0)
+    while (buf->pos < 0 && buf->currentLine->prev && buf->currentLine->bpos) {
+	pos = buf->pos + buf->currentLine->prev->len;
+	cursorUp0(buf, 1);
+	buf->pos = pos;
+    }
+    while (buf->pos >= buf->currentLine->len && buf->currentLine->next &&
+	   buf->currentLine->next->bpos) {
+	pos = buf->pos - buf->currentLine->len;
+	cursorDown0(buf, 1);
+	buf->pos = pos;
+    }
+    if (buf->currentLine->len == 0 || buf->pos < 0)
 	buf->pos = 0;
     else if (buf->pos >= buf->currentLine->len)
 	buf->pos = buf->currentLine->len - 1;
@@ -1206,8 +1473,9 @@ arrangeCursor(Buffer *buf)
     }
     /* Arrange cursor */
     buf->cursorY = buf->currentLine->linenumber - buf->topLine->linenumber;
-    buf->visualpos = buf->cursorX =
+    buf->visualpos = buf->currentLine->bwidth +
 	COLPOS(buf->currentLine, buf->pos) - buf->currentColumn;
+    buf->cursorX = buf->visualpos - buf->currentLine->bwidth;
 #ifdef DISPLAY_DEBUG
     fprintf(stderr,
 	    "arrangeCursor: column=%d, cursorX=%d, visualpos=%d, pos=%d, len=%d\n",
@@ -1224,7 +1492,8 @@ arrangeLine(Buffer *buf)
     if (buf->firstLine == NULL)
 	return;
     buf->cursorY = buf->currentLine->linenumber - buf->topLine->linenumber;
-    i = columnPos(buf->currentLine, buf->currentColumn + buf->visualpos);
+    i = columnPos(buf->currentLine, buf->currentColumn + buf->visualpos
+		  - buf->currentLine->bwidth);
     cpos = COLPOS(buf->currentLine, i) - buf->currentColumn;
     if (cpos >= 0) {
 	buf->cursorX = cpos;
@@ -1286,6 +1555,8 @@ restorePosition(Buffer *buf, Buffer *orig)
 			    FALSE);
     gotoLine(buf, CUR_LINENUMBER(orig));
     buf->pos = orig->pos;
+    if (buf->currentLine && orig->currentLine)
+	buf->pos += orig->currentLine->bpos - buf->currentLine->bpos;
     buf->currentColumn = orig->currentColumn;
     arrangeCursor(buf);
 }
