@@ -1,4 +1,4 @@
-/* $Id: etc.c,v 1.28 2002/10/30 17:21:40 ukai Exp $ */
+/* $Id: etc.c,v 1.66 2003/02/18 15:44:33 ukai Exp $ */
 #include "fm.h"
 #include <pwd.h>
 #include "myctype.h"
@@ -7,11 +7,7 @@
 #include "hash.h"
 #include "terms.h"
 
-#ifdef HAVE_GETCWD		/* ??? ukai */
-#include <unistd.h>
-#include <sys/param.h>
-#endif				/* HAVE_GETCWD */
-
+#include <fcntl.h>
 #include <sys/types.h>
 #include <time.h>
 #if defined(HAVE_WAITPID) || defined(HAVE_WAIT3)
@@ -42,7 +38,7 @@ columnSkip(Buffer *buf, int offset)
 {
     int i, maxColumn;
     int column = buf->currentColumn + offset;
-    int nlines = LASTLINE + 1;
+    int nlines = buf->LINES + 1;
     Line *l;
 
     maxColumn = 0;
@@ -91,7 +87,7 @@ lineSkip(Buffer *buf, Line *line, int offset, int last)
 #ifdef NEXTPAGE_TOPLINE
     if (!nextpage_topline)
 #endif
-	for (i = (LASTLINE - 1) - (buf->lastLine->linenumber - l->linenumber);
+	for (i = buf->LINES - 1 - (buf->lastLine->linenumber - l->linenumber);
 	     i > 0 && l->prev != NULL; i--, l = l->prev) ;
     return l;
 }
@@ -103,7 +99,7 @@ currentLineSkip(Buffer *buf, Line *line, int offset, int last)
     Line *l = line;
 
     if (buf->pagerSource && !(buf->bufferprop & BP_CLOSE)) {
-	n = line->linenumber + offset + LASTLINE;
+	n = line->linenumber + offset + buf->LINES;
 	if (buf->lastLine->linenumber < n)
 	    getNextPage(buf, n - buf->lastLine->linenumber);
 	while ((last || (buf->lastLine->linenumber < n)) &&
@@ -134,14 +130,17 @@ gethtmlcmd(char **s)
 
     (*s)++;
     /* first character */
-    if (IS_ALNUM(**s) || **s == '_' || **s == '/')
-	*(p++) = tolower(*((*s)++));
+    if (IS_ALNUM(**s) || **s == '_' || **s == '/') {
+	*(p++) = TOLOWER(**s);
+	(*s)++;
+    }
     else
 	return HTML_UNKNOWN;
     if (p[-1] == '/')
 	SKIP_BLANKS(*s);
     while ((IS_ALNUM(**s) || **s == '_') && p - cmdstr < MAX_CMD_LEN) {
-	*(p++) = tolower(*((*s)++));
+	*(p++) = TOLOWER(**s);
+	(*s)++;
     }
     if (p - cmdstr == MAX_CMD_LEN) {
 	/* buffer overflow: perhaps caused by bad HTML source */
@@ -237,50 +236,58 @@ parse_ansi_color(char **str, Lineprop *effect, Linecolor *color)
  */
 
 Str
-checkType(Str s, Lineprop *oprop,
-#ifdef USE_ANSI_COLOR
-	  Linecolor *ocolor, int *check_color,
-#endif
-	  int len)
+checkType(Str s, Lineprop **oprop, Linecolor **ocolor)
 {
     Lineprop mode;
     Lineprop effect = PE_NORMAL;
-    Lineprop *prop = oprop;
+    Lineprop *prop;
+    static Lineprop *prop_buffer = NULL;
+    static int prop_size = 0;
     char *str = s->ptr, *endp = &s->ptr[s->length], *bs = NULL;
 #ifdef USE_ANSI_COLOR
     Lineprop ceffect = PE_NORMAL;
     Linecolor cmode = 0;
+    int check_color = FALSE;
     Linecolor *color = NULL;
+    static Linecolor *color_buffer = NULL;
+    static int color_size = 0;
     char *es = NULL;
 #endif
     int do_copy = FALSE;
-    int size = (len < s->length) ? len : s->length;
 
-#ifdef USE_ANSI_COLOR
-    if (check_color)
-	*check_color = FALSE;
-#endif
+    if (prop_size < s->length) {
+	prop_size = (s->length > LINELEN) ? s->length : LINELEN;
+	prop_buffer = New_Reuse(Lineprop, prop_buffer, prop_size);
+    }
+    prop = prop_buffer;
+
     if (ShowEffect) {
 	bs = memchr(str, '\b', s->length);
 #ifdef USE_ANSI_COLOR
 	if (ocolor) {
 	    es = memchr(str, ESC_CODE, s->length);
-	    if (es)
-		color = ocolor;
+	    if (es) {
+		if (color_size < s->length) {
+		    color_size = (s->length > LINELEN) ? s->length : LINELEN;
+		    color_buffer = New_Reuse(Linecolor, color_buffer,
+					     color_size);
+		}
+		color = color_buffer;
+	    }
 	}
 #endif
-	if (s->length > size || (bs != NULL)
+	if ((bs != NULL)
 #ifdef USE_ANSI_COLOR
 	    || (es != NULL)
 #endif
 	    ) {
-	    s = Strnew_size(size);
+	    s = Strnew_size(s->length);
 	    do_copy = TRUE;
 	}
     }
 
     while (str < endp) {
-	if (prop - oprop >= len)
+	if (prop - prop_buffer >= prop_size)
 	    break;
 	if (bs != NULL) {
 	    if (str == bs - 2 && !strncmp(str, "__\b\b", 4)) {
@@ -411,7 +418,7 @@ checkType(Str s, Lineprop *oprop,
 		    es = memchr(str, ESC_CODE, endp - str);
 		if (ok) {
 		    if (cmode)
-			*check_color = TRUE;
+			check_color = TRUE;
 		    continue;
 		}
 	    }
@@ -459,13 +466,35 @@ checkType(Str s, Lineprop *oprop,
 	}
 	effect = PE_NORMAL;
     }
+    *oprop = prop_buffer;
+#ifdef USE_ANSI_COLOR
+    if (ocolor)
+	*ocolor = check_color ? color_buffer : NULL;
+#endif
     return s;
 }
 
+static int
+nextColumn(int n, char *p, Lineprop *pr)
+{
+    if (*p == '\t' && *pr == PC_CTRL)
+	return (n + Tabstop) / Tabstop * Tabstop;
+#ifndef KANJI_SYMBOLS
+    else if (*pr & PC_RULE)
+	return n + 1;
+#endif
+    else if (IS_UNPRINTABLE_ASCII(*p, *pr))
+	return n + 4;
+    else if (IS_UNPRINTABLE_CONTROL(*p, *pr))
+	return n + 2;
+    return n + 1;
+}
+
 int
 calcPosition(char *l, Lineprop *pr, int len, int pos, int bpos, int mode)
 {
-    static short realColumn[LINELEN + 1];
+    static int *realColumn = NULL;
+    static int size = 0;
     static char *prevl = NULL;
     int i, j;
 
@@ -475,30 +504,41 @@ calcPosition(char *l, Lineprop *pr, int len, int pos, int bpos, int mode)
 	if (pos <= len)
 	    return realColumn[pos];
     }
+    if (size < len + 1) {
+	size = (len + 1 > LINELEN) ? (len + 1) : LINELEN;
+	realColumn = New_N(int, size);
+    }
     prevl = l;
     j = bpos;
     for (i = 0;; i++) {
 	realColumn[i] = j;
 	if (i == len)
 	    break;
-	if (l[i] == '\t' && pr[i] == PC_CTRL)
-	    j = (j + Tabstop) / Tabstop * Tabstop;
-#ifndef KANJI_SYMBOLS
-	else if (pr[i] & PC_RULE)
-	    j++;
-#endif
-	else if (IS_UNPRINTABLE_ASCII(l[i], pr[i]))
-	    j = j + 4;
-	else if (IS_UNPRINTABLE_CONTROL(l[i], pr[i]))
-	    j = j + 2;
-	else
-	    j++;
+	j = nextColumn(j, &l[i], &pr[i]);
     }
     if (pos >= i)
 	return j;
     return realColumn[pos];
 }
 
+int
+columnLen(Line *line, int column)
+{
+    int i, j;
+
+    for (i = 0, j = 0; i < line->len; i++) {
+	j = nextColumn(j, &line->lineBuf[i], &line->propBuf[i]);
+	if (j > column) {
+#ifdef JP_CHARSET
+	    if (CharType(line->propBuf[i]) == PC_KANJI2)
+		return i - 1;
+#endif
+	    return i;
+	}
+    }
+    return line->len;
+}
+
 char *
 lastFileName(char *path)
 {
@@ -637,7 +677,7 @@ next_status(char c, int *status)
 	else if (c == '>')
 	    *status = R_ST_NORMAL;
 	else
-	    *status = R_ST_TAG;
+	    *status = R_ST_VALUE;
 	return 0;
     case R_ST_QUOTE:
 	if (c == '\'')
@@ -647,6 +687,12 @@ next_status(char c, int *status)
 	if (c == '"')
 	    *status = R_ST_TAG;
 	return 0;
+    case R_ST_VALUE:
+	if (c == '>')
+	    *status = R_ST_NORMAL;
+	else if (IS_SPACE(c))
+	    *status = R_ST_TAG;
+	return 0;
     case R_ST_AMP:
 	if (c == ';') {
 	    *status = R_ST_NORMAL;
@@ -755,8 +801,10 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
 	    }
 	    if (prev_status == R_ST_NCMNT2 || prev_status == R_ST_NCMNT3 ||
 		prev_status == R_ST_IRRTAG || prev_status == R_ST_CMNT1) {
-		if (prev_status == R_ST_CMNT1 && !append)
+		if (prev_status == R_ST_CMNT1 && !append && !pre)
 		    Strclear(buf);
+		if (pre)
+		    Strcat_char(buf, *p);
 		p++;
 		goto proc_end;
 	    }
@@ -779,7 +827,10 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
 	    }
 	    if (*status == R_ST_TAG0 && !REALLY_THE_BEGINNING_OF_A_TAG(p)) {
 		/* it seems that this '<' is not a beginning of a tag */
-		Strcat_charp(buf, "&lt;");
+		/*
+		 * Strcat_charp(buf, "&lt;");
+		 */
+		Strcat_char(buf, '<');
 		*status = R_ST_NORMAL;
 	    }
 	    else
@@ -788,12 +839,15 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
 	case R_ST_EQL:
 	case R_ST_QUOTE:
 	case R_ST_DQUOTE:
+	case R_ST_VALUE:
 	case R_ST_AMP:
 	    Strcat_char(buf, *p);
 	    break;
 	case R_ST_CMNT:
 	case R_ST_IRRTAG:
-	    if (!append)
+	    if (pre)
+		Strcat_char(buf, *p);
+	    else if (!append)
 		Strclear(buf);
 	    break;
 	case R_ST_CMNT1:
@@ -802,6 +856,8 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
 	case R_ST_NCMNT2:
 	case R_ST_NCMNT3:
 	    /* do nothing */
+	    if (pre)
+		Strcat_char(buf, *p);
 	    break;
 	}
     }
@@ -830,6 +886,7 @@ correct_irrtag(int status)
 	case R_ST_TAG:
 	case R_ST_TAG0:
 	case R_ST_EQL:		/* required ">" */
+	case R_ST_VALUE:
 	    c = '>';
 	    break;
 	case R_ST_QUOTE:
@@ -1033,48 +1090,61 @@ parsePasswd(FILE * fp, int netrc)
     add_auth_pass_entry(&ent, netrc);
 }
 
-#define PASS_IS_READABLE_MSG "SECURITY NOTE: passwd file must not be accessible by others"
+#define FILE_IS_READABLE_MSG "SECURITY NOTE: file %s must not be accessible by others"
 
-static FILE *
-openPasswdFile(char *fname)
+FILE *
+openSecretFile(char *fname)
 {
+    char *efname;
     struct stat st;
+
     if (fname == NULL)
 	return NULL;
-    if (stat(expandName(fname), &st) < 0)
+    efname = expandPath(fname);
+    if (stat(efname, &st) < 0)
 	return NULL;
 
     /* check permissions, if group or others readable or writable,
      * refuse it, because it's insecure.
+     *
+     * XXX: disable_secret_security_check will introduce some
+     *    security issues, but on some platform such as Windows
+     *    it's not possible (or feasible) to disable group|other
+     *    readable and writable.
+     *   [w3m-dev 03368][w3m-dev 03369][w3m-dev 03370]
      */
-    if ((st.st_mode & (S_IRWXG | S_IRWXO)) != 0) {
+    if (disable_secret_security_check)
+	/* do nothing */ ;
+    else if ((st.st_mode & (S_IRWXG | S_IRWXO)) != 0) {
 	if (fmInitialized) {
-	    message(PASS_IS_READABLE_MSG, 0, 0);
+	    message(Sprintf(FILE_IS_READABLE_MSG, fname)->ptr, 0, 0);
 	    refresh();
 	}
 	else {
-	    fputs(PASS_IS_READABLE_MSG, stderr);
+	    fputs(Sprintf(FILE_IS_READABLE_MSG, fname)->ptr, stderr);
 	    fputc('\n', stderr);
 	}
 	sleep(2);
 	return NULL;
     }
 
-    return fopen(expandName(fname), "r");
+    return fopen(efname, "r");
 }
 
 void
 loadPasswd(void)
 {
     FILE *fp;
-    fp = openPasswdFile(passwd_file);
+
+    passwords = NULL;
+    fp = openSecretFile(passwd_file);
     if (fp != NULL) {
 	parsePasswd(fp, 0);
 	fclose(fp);
     }
 
     /* for FTP */
-    fp = openPasswdFile("~/.netrc");
+    fp = openSecretFile("~/.netrc");
     if (fp != NULL) {
 	parsePasswd(fp, 1);
 	fclose(fp);
@@ -1262,53 +1332,143 @@ romanAlphabet(int n)
 #define SIGIOT SIGABRT
 #endif				/* not SIGIOT */
 
-void
+static void
 reset_signals(void)
 {
-    signal(SIGHUP, SIG_DFL);	/* terminate process */
-    signal(SIGINT, SIG_DFL);	/* terminate process */
-    signal(SIGQUIT, SIG_DFL);	/* terminate process */
-    signal(SIGTERM, SIG_DFL);	/* terminate process */
-    signal(SIGILL, SIG_DFL);	/* create core image */
-    signal(SIGIOT, SIG_DFL);	/* create core image */
-    signal(SIGFPE, SIG_DFL);	/* create core image */
+    mySignal(SIGHUP, SIG_DFL);	/* terminate process */
+    mySignal(SIGINT, SIG_DFL);	/* terminate process */
+    mySignal(SIGQUIT, SIG_DFL);	/* terminate process */
+    mySignal(SIGTERM, SIG_DFL);	/* terminate process */
+    mySignal(SIGILL, SIG_DFL);	/* create core image */
+    mySignal(SIGIOT, SIG_DFL);	/* create core image */
+    mySignal(SIGFPE, SIG_DFL);	/* create core image */
 #ifdef SIGBUS
-    signal(SIGBUS, SIG_DFL);	/* create core image */
+    mySignal(SIGBUS, SIG_DFL);	/* create core image */
 #endif				/* SIGBUS */
 #ifdef SIGCHLD
-    signal(SIGCHLD, SIG_IGN);
+    mySignal(SIGCHLD, SIG_IGN);
 #endif
 #ifdef SIGPIPE
-    signal(SIGPIPE, SIG_IGN);
+    mySignal(SIGPIPE, SIG_IGN);
 #endif
-    signal(SIGUSR1, SIG_IGN);
+}
+
+#ifndef FOPEN_MAX
+#define FOPEN_MAX 1024		/* XXX */
+#endif
+
+static void
+close_all_fds_except(int i, int f)
+{
+    switch (i) {		/* fall through */
+    case 0:
+	dup2(open(DEV_NULL_PATH, O_RDONLY), 0);
+    case 1:
+	dup2(open(DEV_NULL_PATH, O_WRONLY), 1);
+    case 2:
+	dup2(open(DEV_NULL_PATH, O_WRONLY), 2);
+    }
+    /* close all other file descriptors (socket, ...) */
+    for (i = 3; i < FOPEN_MAX; i++) {
+	if (i != f)
+	    close(i);
+    }
+}
+
+void
+setup_child(int child, int i, int f)
+{
+    reset_signals();
+    mySignal(SIGINT, SIG_IGN);
+    if (!child)
+	SETPGRP();
+    close_tty();
+    close_all_fds_except(i, f);
+    QuietMessage = TRUE;
+    fmInitialized = FALSE;
+    TrapSignal = FALSE;
+}
+
+pid_t
+open_pipe_rw(FILE ** fr, FILE ** fw)
+{
+    int fdr[2];
+    int fdw[2];
+    pid_t pid;
+
+    if (fr && pipe(fdr) < 0)
+	goto err0;
+    if (fw && pipe(fdw) < 0)
+	goto err1;
+
+    flush_tty();
+    pid = fork();
+    if (pid < 0)
+	goto err2;
+    if (pid == 0) {
+	/* child */
+	if (fr) {
+	    close(fdr[0]);
+	    dup2(fdr[1], 1);
+	}
+	if (fw) {
+	    close(fdw[1]);
+	    dup2(fdw[0], 0);
+	}
+    }
+    else {
+	if (fr) {
+	    close(fdr[1]);
+	    if (*fr == stdin)
+		dup2(fdr[0], 0);
+	    else
+		*fr = fdopen(fdr[0], "r");
+	}
+	if (fw) {
+	    close(fdw[0]);
+	    if (*fw == stdout)
+		dup2(fdw[1], 1);
+	    else
+		*fw = fdopen(fdw[1], "w");
+	}
+    }
+    return pid;
+  err2:
+    if (fw) {
+	close(fdw[0]);
+	close(fdw[1]);
+    }
+  err1:
+    if (fr) {
+	close(fdr[0]);
+	close(fdr[1]);
+    }
+  err0:
+    return (pid_t) - 1;
+}
+
+void
+myExec(char *command)
+{
+    mySignal(SIGINT, SIG_DFL);
+    execl("/bin/sh", "sh", "-c", command, NULL);
+    exit(127);
 }
 
 void
 mySystem(char *command, int background)
 {
     if (background) {
-#ifndef HAVE_SETPGRP
+#ifndef __EMX__
+	flush_tty();
+	if (!fork()) {
+	    setup_child(FALSE, 0, -1);
+	    myExec(command);
+	}
+#else
 	Str cmd = Strnew_charp("start /f ");
 	Strcat_charp(cmd, command);
 	system(cmd->ptr);
-#else
-	int pid;
-	flush_tty();
-	if ((pid = fork()) == 0) {
-	    int fd, i;
-	    reset_signals();
-	    SETPGRP();
-	    close_tty();
-	    dup2(open("/dev/null", O_RDONLY), 0);
-	    dup2(open("/dev/null", O_WRONLY), 1);
-	    dup2(fd = open("/dev/null", O_WRONLY), 2);
-	    /* close all other file descriptors (socket, ...) */
-	    for (i = 3; i <= fd; i++)
-		close(i);
-	    execl("/bin/sh", "sh", "-c", command, NULL);
-	    exit(127);
-	}
 #endif
     }
     else
@@ -1380,35 +1540,44 @@ myEditor(char *cmd, char *file, int line)
 char *
 expandName(char *name)
 {
-    Str userName = NULL;
     char *p;
     struct passwd *passent, *getpwnam(const char *);
-    Str extpath = Strnew();
+    Str extpath = NULL;
 
+    if (name == NULL)
+	return NULL;
     p = name;
-    if (*p == '/' && *(p + 1) == '~' && IS_ALPHA(*(p + 2))) {
-	if (personal_document_root != NULL) {
-	    userName = Strnew();
+    if (*p == '/') {
+	if (*(p + 1) == '~' && IS_ALPHA(*(p + 2)) && personal_document_root) {
+	    char *q;
 	    p += 2;
-	    while (IS_ALNUM(*p) || *p == '_' || *p == '-')
-		Strcat_char(userName, *(p++));
-	    passent = getpwnam(userName->ptr);
-	    if (passent == NULL) {
-		p = name;
-		goto rest;
+	    q = strchr(p, '/');
+	    if (q) {		/* /~user/dir... */
+		passent = getpwnam(allocStr(p, q - p));
+		p = q;
 	    }
-	    Strcat_charp(extpath, passent->pw_dir);
-	    Strcat_char(extpath, '/');
-	    Strcat_charp(extpath, personal_document_root);
-	    if (Strcmp_charp(extpath, "/") == 0 && *p == '/')
+	    else {		/* /~user */
+		passent = getpwnam(p);
+		p = "";
+	    }
+	    if (!passent)
+		goto rest;
+	    extpath = Strnew_m_charp(passent->pw_dir, "/",
+				     personal_document_root, NULL);
+	    if (*personal_document_root == '\0' && *p == '/')
 		p++;
 	}
+	else
+	    goto rest;
+	if (Strcmp_charp(extpath, "/") == 0 && *p == '/')
+	    p++;
+	Strcat_charp(extpath, p);
+	return extpath->ptr;
     }
     else
-	p = expandPath(p);
+	return expandPath(p);
   rest:
-    Strcat_charp(extpath, p);
-    return extpath->ptr;
+    return name;
 }
 
 char *
@@ -1422,7 +1591,7 @@ file_to_url(char *file)
     char *host = NULL;
 #endif
 
-    file = expandName(file);
+    file = expandPath(file);
 #ifdef SUPPORT_NETBIOS_SHARE
     if (file[0] == '/' && file[1] == '/') {
 	char *p;
@@ -1463,8 +1632,21 @@ file_to_url(char *file)
     return tmp->ptr;
 }
 
+char *
+url_unquote_conv(char *url, char code)
+{
+    Str tmp;
+    tmp = Str_url_unquote(Strnew_charp(url), FALSE, TRUE);
+#ifdef JP_CHARSET
+    if (code == CODE_INNER_EUC)
+	code = CODE_EUC;
+    tmp = convertLine(NULL, tmp, &code, RAW_MODE);
+#endif
+    return tmp->ptr;
+}
+
 static char *tmpf_base[MAX_TMPF_TYPE] = {
-    "tmp", "src", "frame", "cache"
+    "tmp", "src", "frame", "cache", "cookie",
 };
 static unsigned int tmpf_seq[MAX_TMPF_TYPE];
 
@@ -1473,13 +1655,13 @@ tmpfname(int type, char *ext)
 {
     Str tmpf;
     tmpf = Sprintf("%s/w3m%s%d-%d%s",
-		   rc_dir,
+		   tmp_dir,
 		   tmpf_base[type],
-		   (int)getpid(), tmpf_seq[type]++, (ext) ? ext : "");
+		   CurrentPid, tmpf_seq[type]++, (ext) ? ext : "");
+    pushText(fileToDelete, tmpf->ptr);
     return tmpf;
 }
 
-#ifdef USE_COOKIE
 static char *monthtbl[] = {
     "Jan", "Feb", "Mar", "Apr", "May", "Jun",
     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
@@ -1605,12 +1787,38 @@ get_time(char **s, int *hour, int *min, int *sec)
     return 0;
 }
 
+static int
+get_zone(char **s, int *z_hour, int *z_min)
+{
+    Str tmp = Strnew();
+    int zone;
+    char *ss = *s;
+
+    if (!**s)
+	return -1;
+
+    if (**s == '+' || **s == '-')
+	Strcat_char(tmp, *((*s)++));
+    while (**s && IS_DIGIT(**s))
+	Strcat_char(tmp, *((*s)++));
+    if (!(tmp->length == 4 && IS_DIGIT(*ss)) &&
+	!(tmp->length == 5 && (*ss == '+' || *ss == '-'))) {
+	*s = ss;
+	return -1;
+    }
+
+    zone = atoi(tmp->ptr);
+    *z_hour = zone / 100;
+    *z_min = zone - (zone / 100) * 100;
+    return 0;
+}
+
 /* RFC 1123 or RFC 850 or ANSI C asctime() format string -> time_t */
 time_t
 mymktime(char *timestr)
 {
     char *s;
-    int day, mon, year, hour, min, sec;
+    int day, mon, year, hour, min, sec, z_hour = 0, z_min = 0;
 
     if (!(timestr && *timestr))
 	return -1;
@@ -1647,8 +1855,12 @@ mymktime(char *timestr)
 	    min = 0;
 	    sec = 0;
 	}
-	else if (get_time(&s, &hour, &min, &sec) == -1) {
-	    return -1;
+	else {
+	    if (get_time(&s, &hour, &min, &sec) == -1)
+		return -1;
+	    while (*s && !IS_DIGIT(*s) && *s != '+' && *s != '-')
+		s++;
+	    get_zone(&s, &z_hour, &z_min);
 	}
     }
     else {
@@ -1674,8 +1886,9 @@ mymktime(char *timestr)
 	    return -1;
     }
 #ifdef DEBUG
-    fprintf(stderr, "year=%d month=%d day=%d hour:min:sec=%d:%d:%d\n",
-	    year, mon, day, hour, min, sec);
+    fprintf(stderr,
+	    "year=%d month=%d day=%d hour:min:sec=%d:%d:%d zone=%d:%d\n", year,
+	    mon, day, hour, min, sec, z_hour, z_min);
 #endif				/* DEBUG */
 
     mon -= 3;
@@ -1685,10 +1898,13 @@ mymktime(char *timestr)
     }
     day += (year - 1968) * 1461 / 4;
     day += ((((mon * 153) + 2) / 5) - 672);
+    hour -= z_hour;
+    min -= z_min;
     return (time_t) ((day * 60 * 60 * 24) +
 		     (hour * 60 * 60) + (min * 60) + sec);
 }
 
+#ifdef USE_COOKIE
 #ifdef INET6
 #include <sys/socket.h>
 #endif				/* INET6 */
@@ -1757,3 +1973,26 @@ FQDN(char *host)
 }
 
 #endif				/* USE_COOKIE */
+
+void (*mySignal(int signal_number, void (*action) (int))) (int) {
+#ifdef	SA_RESTART
+    struct sigaction new_action, old_action;
+
+    sigemptyset(&new_action.sa_mask);
+    new_action.sa_handler = action;
+    if (signal_number == SIGALRM) {
+#ifdef	SA_INTERRUPT
+	new_action.sa_flags = SA_INTERRUPT;
+#else
+	new_action.sa_flags = 0;
+#endif
+    }
+    else {
+	new_action.sa_flags = SA_RESTART;
+    }
+    sigaction(signal_number, &new_action, &old_action);
+    return (old_action.sa_handler);
+#else
+    return (signal(signal_number, action));
+#endif
+}
