@@ -1,4 +1,4 @@
-/* $Id: form.c,v 1.14 2002/06/17 17:51:39 ukai Exp $ */
+/* $Id: form.c,v 1.30 2003/01/29 17:10:38 ukai Exp $ */
 /* 
  * HTML forms
  */
@@ -7,13 +7,7 @@
 #include "parsetagx.h"
 #include "myctype.h"
 #include "local.h"
-
-#ifndef HAVE_LSTAT
-/* lstat is identical to stat, only the link itself is statted, not the file
- * that is obtained by tracing the links. But on OS/2 systems, there is no
- * differences. */
-#define lstat stat
-#endif				/* not HAVE_LSTAT */
+#include "regex.h"
 
 extern Str *textarea_str;
 #ifdef MENU_SELECT
@@ -31,6 +25,7 @@ struct {
 #ifdef USE_COOKIE
     {"cookie", set_cookie_flag},
 #endif				/* USE_COOKIE */
+    {"download", download_action},
     {"none", NULL},
     {NULL, NULL},
 };
@@ -158,8 +153,8 @@ formList_addInput(struct form_list *fl, struct parsed_tag *tag)
 }
 
 static char *_formtypetbl[] = {
-    "text", "password", "checkbox", "radio", "submit",
-    "reset", "hidden", "image", "select", "textarea", "button", "file", 0,
+    "text", "password", "checkbox", "radio", "submit", "reset", "hidden",
+    "image", "select", "textarea", "button", "file", NULL
 };
 
 static char *_formmethodtbl[] = {
@@ -169,16 +164,19 @@ static char *_formmethodtbl[] = {
 char *
 form2str(FormItemList *fi)
 {
-    Str tmp;
-    if (fi->type == FORM_INPUT_SUBMIT ||
-	fi->type == FORM_INPUT_IMAGE || fi->type == FORM_INPUT_BUTTON) {
-	tmp = Strnew_charp(_formmethodtbl[fi->parent->method]);
-	Strcat_char(tmp, ' ');
-	Strcat(tmp, fi->parent->action);
-	return tmp->ptr;
-    }
-    else
-	return _formtypetbl[fi->type];
+    Str tmp = Strnew();
+
+    if (fi->type != FORM_SELECT && fi->type != FORM_TEXTAREA)
+	Strcat_charp(tmp, "input type=");
+    Strcat_charp(tmp, _formtypetbl[fi->type]);
+    if (fi->name && fi->name->length)
+	Strcat_m_charp(tmp, " name=\"", fi->name->ptr, "\"", NULL);
+    if ((fi->type == FORM_INPUT_RADIO || fi->type == FORM_INPUT_CHECKBOX ||
+	 fi->type == FORM_SELECT) && fi->value)
+	Strcat_m_charp(tmp, " value=\"", fi->value->ptr, "\"", NULL);
+    Strcat_m_charp(tmp, " (", _formmethodtbl[fi->parent->method], " ",
+		   fi->parent->action->ptr, ")", NULL);
+    return tmp->ptr;
 }
 
 int
@@ -429,16 +427,6 @@ form_fputs_decode(Str s, FILE * f)
 
     for (p = s->ptr; *p;) {
 	switch (*p) {
-	case '<':
-	    if (!strncasecmp(p, "<eol>", 5)) {
-		Strcat_char(z, '\n');
-		p += 5;
-	    }
-	    else {
-		Strcat_char(z, *p);
-		p++;
-	    }
-	    break;
 #if !defined( __CYGWIN__ ) && !defined( __EMX__ )
 	case '\r':
 	    if (*(p + 1) == '\n')
@@ -452,11 +440,8 @@ form_fputs_decode(Str s, FILE * f)
 	    break;
 	}
     }
-#ifdef JP_CHARSET
-    fputs(conv_str(z, InnerCode, DisplayCode)->ptr, f);
-#else				/* not JP_CHARSET */
-    fputs(z->ptr, f);
-#endif				/* not JP_CHARSET */
+    z = conv_str(z, InnerCode, DisplayCode);
+    Strfputs(z, f);
 }
 
 
@@ -467,7 +452,7 @@ input_textarea(FormItemList *fi)
     Str tmp;
     FILE *f;
 #ifdef JP_CHARSET
-    char code = DisplayCode, ic;
+    char code = DisplayCode;
 #endif
 
     f = fopen(tmpf, "w");
@@ -484,11 +469,11 @@ input_textarea(FormItemList *fi)
     fmInit();
 
     if (fi->readonly)
-	return;
+	goto input_end;
     f = fopen(tmpf, "r");
     if (f == NULL) {
 	disp_err_message("Can't open temporary file", FALSE);
-	return;
+	goto input_end;
     }
     fi->value = Strnew();
     while (tmp = Strfgets(f), tmp->length > 0) {
@@ -502,12 +487,12 @@ input_textarea(FormItemList *fi)
 	    Strcat_charp(tmp, "\r\n");
 	}
 #ifdef JP_CHARSET
-	if ((ic = checkShiftCode(tmp, code)) != '\0')
-	    tmp = conv_str(tmp, (code = ic), InnerCode);
+	tmp = convertLine(NULL, tmp, &code, RAW_MODE);
 #endif				/* not JP_CHARSET */
 	Strcat(fi->value, tmp);
     }
     fclose(f);
+  input_end:
     unlink(tmpf);
 }
 
@@ -655,3 +640,287 @@ form_write_from_file(FILE * f, char *boundary, char *name, char *filename,
   write_end:
     fprintf(f, "\r\n");
 }
+
+struct pre_form_item {
+    int type;
+    char *name;
+    char *value;
+    int checked;
+    struct pre_form_item *next;
+};
+
+struct pre_form {
+    char *url;
+    Regex *re_url;
+    char *name;
+    char *action;
+    struct pre_form_item *item;
+    struct pre_form *next;
+};
+
+static struct pre_form *PreForm = NULL;
+
+static struct pre_form *
+add_pre_form(struct pre_form *prev, char *url, char *name, char *action)
+{
+    ParsedURL pu;
+    struct pre_form *new;
+
+    if (prev)
+	new = prev->next = New(struct pre_form);
+    else
+	new = PreForm = New(struct pre_form);
+    if (url && *url == '/') {
+	int l = strlen(url);
+	if (l > 1 && url[l - 1] == '/')
+	    new->url = allocStr(url + 1, l - 2);
+	else
+	    new->url = url + 1;
+	new->re_url = newRegex(new->url, FALSE, NULL, NULL);
+	if (!new->re_url)
+	    new->url = NULL;
+    }
+    else if (url) {
+	parseURL2(url, &pu, NULL);
+	new->url = parsedURL2Str(&pu)->ptr;
+	new->re_url = NULL;
+    }
+    new->name = (name && *name) ? name : NULL;
+    new->action = (action && *action) ? action : NULL;
+    new->item = NULL;
+    new->next = NULL;
+    return new;
+}
+
+static struct pre_form_item *
+add_pre_form_item(struct pre_form *pf, struct pre_form_item *prev, int type,
+		  char *name, char *value, char *checked)
+{
+    struct pre_form_item *new;
+
+    if (!pf)
+	return NULL;
+    if (prev)
+	new = prev->next = New(struct pre_form_item);
+    else
+	new = pf->item = New(struct pre_form_item);
+    new->type = type;
+    new->name = name;
+    new->value = value;
+    if (checked && *checked && (!strcmp(checked, "0") ||
+				strcasecmp(checked, "off")
+				|| !strcasecmp(checked, "no")))
+	new->checked = 0;
+    else
+	new->checked = 1;
+    new->next = NULL;
+    return new;
+}
+
+/*
+ * url <url>|/<re-url>/
+ * form [<name>] <action>
+ * text <name> <value>
+ * file <name> <value>
+ * passwd <name> <value>
+ * checkbox <name> <value> [<checked>]
+ * radio <name> <value>
+ * select <name> <value>
+ * submit [<name> [<value>]]
+ * textarea <name>
+ * <value>
+ * /textarea
+ */
+
+void
+loadPreForm(void)
+{
+    FILE *fp;
+    Str line = NULL, textarea = NULL;
+    struct pre_form *pf = NULL;
+    struct pre_form_item *pi = NULL;
+    int type = -1;
+    char *name = NULL;
+
+    PreForm = NULL;
+    fp = openSecretFile(pre_form_file);
+    if (fp == NULL)
+	return;
+    while (1) {
+	char *p, *s, *arg;
+
+	line = Strfgets(fp);
+	if (line->length == 0)
+	    break;
+	if (textarea && !(!strncmp(line->ptr, "/textarea", 9) &&
+			  IS_SPACE(line->ptr[9]))) {
+	    Strcat(textarea, line);
+	    continue;
+	}
+	Strchop(line);
+	Strremovefirstspaces(line);
+	p = line->ptr;
+	if (*p == '#' || *p == '\0')
+	    continue;		/* comment or empty line */
+	s = getWord(&p);
+	arg = getWord(&p);
+
+	if (!strcmp(s, "url")) {
+	    if (!arg || !*arg)
+		continue;
+	    p = getQWord(&p);
+	    pf = add_pre_form(pf, arg, NULL, p);
+	    pi = pf->item;
+	    continue;
+	}
+	if (!pf)
+	    continue;
+	if (!strcmp(s, "form")) {
+	    if (!arg || !*arg)
+		continue;
+	    s = getQWord(&p);
+	    p = getQWord(&p);
+	    if (!p || !*p) {
+		p = s;
+		s = NULL;
+	    }
+	    if (pf->item) {
+		struct pre_form *prev = pf;
+		pf = add_pre_form(prev, "", s, p);
+		/* copy previous URL */
+		pf->url = prev->url;
+		pf->re_url = prev->re_url;
+	    }
+	    else {
+		pf->name = s;
+		pf->action = (p && *p) ? p : NULL;
+	    }
+	    pi = pf->item;
+	    continue;
+	}
+	if (!strcmp(s, "text"))
+	    type = FORM_INPUT_TEXT;
+	else if (!strcmp(s, "file"))
+	    type = FORM_INPUT_FILE;
+	else if (!strcmp(s, "passwd") || !strcmp(s, "password"))
+	    type = FORM_INPUT_PASSWORD;
+	else if (!strcmp(s, "checkbox"))
+	    type = FORM_INPUT_CHECKBOX;
+	else if (!strcmp(s, "radio"))
+	    type = FORM_INPUT_RADIO;
+	else if (!strcmp(s, "submit"))
+	    type = FORM_INPUT_SUBMIT;
+	else if (!strcmp(s, "select"))
+	    type = FORM_SELECT;
+	else if (!strcmp(s, "textarea")) {
+	    type = FORM_TEXTAREA;
+	    name = Strnew_charp(arg)->ptr;
+	    textarea = Strnew();
+	    continue;
+	}
+	else if (textarea && name && !strcmp(s, "/textarea")) {
+	    pi = add_pre_form_item(pf, pi, type, name, textarea->ptr, NULL);
+	    textarea = NULL;
+	    name = NULL;
+	    continue;
+	}
+	else
+	    continue;
+	s = getQWord(&p);
+	pi = add_pre_form_item(pf, pi, type, arg, s, getQWord(&p));
+    }
+    fclose(fp);
+}
+
+void
+preFormUpdateBuffer(Buffer *buf)
+{
+    struct pre_form *pf;
+    struct pre_form_item *pi;
+    int i;
+    Anchor *a;
+    FormList *fl;
+    FormItemList *fi;
+#ifdef MENU_SELECT
+    FormSelectOptionItem *opt;
+    int j;
+#endif
+
+    if (!buf || !buf->formitem || !PreForm)
+	return;
+
+    for (pf = PreForm; pf; pf = pf->next) {
+	if (pf->re_url) {
+	    Str url = parsedURL2Str(&buf->currentURL);
+	    if (!RegexMatch(pf->re_url, url->ptr, url->length, 1))
+		continue;
+	}
+	else if (pf->url) {
+	    if (Strcmp_charp(parsedURL2Str(&buf->currentURL), pf->url))
+		continue;
+	}
+	else
+	    continue;
+	for (i = 0; i < buf->formitem->nanchor; i++) {
+	    a = &buf->formitem->anchors[i];
+	    fi = (FormItemList *)a->url;
+	    fl = fi->parent;
+	    if (pf->name && (!fl->name || strcmp(fl->name, pf->name)))
+		continue;
+	    if (pf->action
+		&& (!fl->action || Strcmp_charp(fl->action, pf->action)))
+		continue;
+	    for (pi = pf->item; pi; pi = pi->next) {
+		if (pi->type != fi->type)
+		    continue;
+		if (pi->type == FORM_INPUT_SUBMIT) {
+		    if ((!pi->name || !*pi->name ||
+			 (fi->name && !Strcmp_charp(fi->name, pi->name))) &&
+			(!pi->value || !*pi->value ||
+			 (fi->value && !Strcmp_charp(fi->value, pi->value))))
+			buf->submit = a;
+		    continue;
+		}
+		if (!pi->name || !fi->name || Strcmp_charp(fi->name, pi->name))
+		    continue;
+		switch (pi->type) {
+		case FORM_INPUT_TEXT:
+		case FORM_INPUT_FILE:
+		case FORM_INPUT_PASSWORD:
+		case FORM_TEXTAREA:
+		    fi->value = Strnew_charp(pi->value);
+		    formUpdateBuffer(a, buf, fi);
+		    break;
+		case FORM_INPUT_CHECKBOX:
+		    if (pi->value && fi->value &&
+			!Strcmp_charp(fi->value, pi->value)) {
+			fi->checked = pi->checked;
+			formUpdateBuffer(a, buf, fi);
+		    }
+		    break;
+		case FORM_INPUT_RADIO:
+		    if (pi->value && fi->value &&
+			!Strcmp_charp(fi->value, pi->value))
+			formRecheckRadio(a, buf, fi);
+		    break;
+#ifdef MENU_SELECT
+		case FORM_SELECT:
+		    for (j = 0, opt = fi->select_option; opt != NULL;
+			 j++, opt = opt->next) {
+			if (pi->value && opt->value &&
+			    !Strcmp_charp(opt->value, pi->value)) {
+			    fi->selected = j;
+			    fi->value = opt->value;
+			    fi->label = opt->label;
+			    updateSelectOption(fi, fi->select_option);
+			    formUpdateBuffer(a, buf, fi);
+			    break;
+			}
+		    }
+		    break;
+#endif
+		}
+	    }
+	}
+    }
+}
