@@ -1,4 +1,4 @@
-/* $Id: html.h,v 1.9 2002/10/10 16:59:33 ukai Exp $ */
+/* $Id: html.h,v 1.22 2003/01/15 16:11:43 ukai Exp $ */
 #ifndef _HTML_H
 #define _HTML_H
 #ifdef USE_SSL
@@ -14,7 +14,8 @@
 #define UFgetc(f) ISgetc((f)->stream)
 #define UFundogetc(f) ISundogetc((f)->stream)
 #define UFread(f,buf,len) ISread((f)->stream,buf,len)
-#define UFclose(f) (ISclose((f)->stream), (f)->stream = NULL)
+#define UFclose(f) (void)(ISclose((f)->stream) == 0 && ((f)->stream = NULL))
+#define UFfileno(f) ISfileno((f)->stream)
 
 struct cmdtable {
     char *cmdname;
@@ -72,6 +73,8 @@ typedef struct {
 #ifdef USE_SSL
     char *ssl_certificate;
 #endif
+    char *url;
+    time_t modtime;
 } URLFile;
 
 #define CMP_NOCOMPRESS   0
@@ -189,10 +192,11 @@ typedef struct {
 #define HTML_N_SUP	101
 #define HTML_SUB	102
 #define HTML_N_SUB	103
+#define HTML_LINK       104
+#define HTML_S          105
+#define HTML_N_S        106
 
    /* pseudo tag */
-#define HTML_INTERNAL       106
-#define HTML_N_INTERNAL     107
 #define HTML_SELECT_INT     108
 #define HTML_N_SELECT_INT   109
 #define HTML_OPTION_INT     110
@@ -211,10 +215,15 @@ typedef struct {
 #define HTML_N_INPUT_ALT    123
 #define HTML_IMG_ALT        124
 #define HTML_N_IMG_ALT      125
-#define HTML_EOL            126
-#define HTML_NOP	    127
+#define HTML_NOP	    126
+#define HTML_PRE_PLAIN	    127
+#define HTML_N_PRE_PLAIN    128
+#define HTML_INTERNAL       129
+#define HTML_N_INTERNAL     130
+#define HTML_DIV_INT        131
+#define HTML_N_DIV_INT      132
 
-#define MAX_HTMLTAG	    128
+#define MAX_HTMLTAG	    133
 
 /* Tag attribute */
 
@@ -265,25 +274,28 @@ typedef struct {
 #define ATTR_SHAPE		44
 #define ATTR_COORDS		45
 #define ATTR_ISMAP		46
+#define ATTR_REL		47
+#define ATTR_REV		48
+#define ATTR_TITLE		49
+#define ATTR_ACCESSKEY		50
 
 /* Internal attribute */
-#define ATTR_XOFFSET		49
-#define ATTR_YOFFSET		50
-#define ATTR_TOP_MARGIN		51
-#define ATTR_BOTTOM_MARGIN	52
-#define ATTR_TID		53
-#define ATTR_FID		54
-#define ATTR_FOR_TABLE		55
-#define ATTR_FRAMENAME		56
-#define ATTR_HBORDER		57
-#define ATTR_HSEQ		58
-#define ATTR_NO_EFFECT		59
-#define ATTR_REFERER		60
-#define ATTR_SELECTNUMBER	61
-#define ATTR_TEXTAREANUMBER	62
-#define ATTR_TITLE		63
+#define ATTR_XOFFSET		60
+#define ATTR_YOFFSET		61
+#define ATTR_TOP_MARGIN		62
+#define ATTR_BOTTOM_MARGIN	63
+#define ATTR_TID		64
+#define ATTR_FID		65
+#define ATTR_FOR_TABLE		66
+#define ATTR_FRAMENAME		67
+#define ATTR_HBORDER		68
+#define ATTR_HSEQ		69
+#define ATTR_NO_EFFECT		70
+#define ATTR_REFERER		71
+#define ATTR_SELECTNUMBER	72
+#define ATTR_TEXTAREANUMBER	73
 
-#define MAX_TAGATTR		64
+#define MAX_TAGATTR		74
 
 /* HTML Tag Information Table */
 
@@ -350,10 +362,13 @@ struct environment {
 #define SCM_LOCAL_CGI	5
 #define SCM_EXEC	6
 #define SCM_NNTP	7
-#define SCM_NEWS	8
-#define SCM_MAILTO      9
+#define SCM_NNTP_GROUP	8
+#define SCM_NEWS	9
+#define SCM_NEWS_GROUP	10
+#define SCM_DATA	11
+#define SCM_MAILTO      12
 #ifdef USE_SSL
-#define SCM_HTTPS       10
+#define SCM_HTTPS       13
 #endif				/* USE_SSL */
 
 #endif				/* _HTML_H */
