@@ -1,5 +1,6 @@
-/* $Id: istream.c,v 1.15 2002/09/28 16:30:07 ukai Exp $ */
+/* $Id: istream.c,v 1.21 2003/02/18 15:43:25 ukai Exp $ */
 #include "fm.h"
+#include "myctype.h"
 #include "istream.h"
 #include <signal.h>
 #ifdef USE_SSL
@@ -175,15 +176,17 @@ newEncodedStream(InputStream is, char encoding)
     return stream;
 }
 
-void
+int
 ISclose(InputStream stream)
 {
     MySignalHandler(*prevtrap) ();
-    if (stream == NULL || stream->base.close == NULL)
-	return;
-    prevtrap = signal(SIGINT, SIG_IGN);
+    if (stream == NULL || stream->base.close == NULL ||
+	stream->base.type & IST_UNCLOSE)
+	return -1;
+    prevtrap = mySignal(SIGINT, SIG_IGN);
     stream->base.close(stream->base.handle);
-    signal(SIGINT, prevtrap);
+    mySignal(SIGINT, prevtrap);
+    return 0;
 }
 
 int
@@ -335,11 +338,15 @@ ISfileno(InputStream stream)
 {
     if (stream == NULL)
 	return -1;
-    switch (IStype(stream)) {
+    switch (IStype(stream) & ~IST_UNCLOSE) {
     case IST_BASIC:
 	return *(int *)stream->base.handle;
     case IST_FILE:
 	return fileno(stream->file.handle->f);
+#ifdef USE_SSL
+    case IST_SSL:
+	return stream->ssl.handle->sock;
+#endif
     case IST_ENCODED:
 	return ISfileno(stream->ens.handle->is);
     default:
@@ -443,7 +450,7 @@ ssl_check_cert_ident(X509 * x, char *hostname)
 		}
 	    }
 	    method = X509V3_EXT_get(ex);
-	    method->ext_free(alt);
+	    sk_GENERAL_NAME_free(alt);
 	    if (i < n)		/* Found a match */
 		match_ident = TRUE;
 	    else if (seen_dnsname)
@@ -494,7 +501,7 @@ ssl_get_certificate(SSL * ssl, char *hostname)
 	    emsg = Strnew_charp("No SSL peer certificate: accept? (y/n)");
 	    ans = inputAnswer(emsg->ptr);
 	}
-	if (ans && tolower(*ans) == 'y')
+	if (ans && TOLOWER(*ans) == 'y')
 	    amsg = Strnew_charp
 		("Accept SSL session without any peer certificate");
 	else {
@@ -527,7 +534,7 @@ ssl_get_certificate(SSL * ssl, char *hostname)
 		emsg = Sprintf("%s: accept? (y/n)", em);
 		ans = inputAnswer(emsg->ptr);
 	    }
-	    if (ans && tolower(*ans) == 'y') {
+	    if (ans && TOLOWER(*ans) == 'y') {
 		amsg = Sprintf("Accept unsecure SSL session: "
 			       "unverified: %s", em);
 	    }
@@ -553,7 +560,7 @@ ssl_get_certificate(SSL * ssl, char *hostname)
 	    Strcat_charp(ep, ": accept? (y/n)");
 	    ans = inputAnswer(ep->ptr);
 	}
-	if (ans && tolower(*ans) == 'y') {
+	if (ans && TOLOWER(*ans) == 'y') {
 	    amsg = Strnew_charp("Accept unsecure SSL session:");
 	    Strcat(amsg, emsg);
 	}
@@ -577,7 +584,7 @@ ssl_get_certificate(SSL * ssl, char *hostname)
 	Strcat_m_charp(s, " subject=", buf, NULL);
     xn = X509_get_issuer_name(x);
     if (X509_NAME_get_text_by_NID(xn, NID_commonName, buf, sizeof(buf)) == -1)
-	Strcat_charp(s, ": issuer=<unnown>");
+	Strcat_charp(s, ": issuer=<unknown>");
     else
 	Strcat_m_charp(s, ": issuer=", buf, NULL);
     Strcat_charp(s, "\n\n");
