@@ -1,4 +1,4 @@
-/* $Id: linein.c,v 1.23 2002/06/01 16:31:55 ukai Exp $ */
+/* $Id: linein.c,v 1.27 2003/02/05 16:43:58 ukai Exp $ */
 #include "fm.h"
 #include "local.h"
 #include "myctype.h"
@@ -35,7 +35,7 @@ static int NCFileOffset;
 static void insertself(char c),
 _mvR(void), _mvL(void), _mvRw(void), _mvLw(void), delC(void), insC(void),
 _mvB(void), _mvE(void), _enter(void), _quo(void), _bs(void), _bsw(void),
-killn(void), killb(void), _inbrk(void), _esc(void),
+killn(void), killb(void), _inbrk(void), _esc(void), _editor(void),
 _prev(void), _next(void), _compl(void), _tcompl(void),
 _dcompl(void), _rdcompl(void), _rcompl(void);
 #ifdef __EMX__
@@ -54,7 +54,7 @@ void (*InputKeymap[32]) () = {
 /*  C-@     C-a     C-b     C-c     C-d     C-e     C-f     C-g     */
     _compl, _mvB,   _mvL,   _inbrk, delC,   _mvE,   _mvR,   _inbrk,
 /*  C-h     C-i     C-j     C-k     C-l     C-m     C-n     C-o     */
-    _bs,    iself,  _enter, killn,  iself,  _enter, _next,  iself,
+    _bs,    iself,  _enter, killn,  iself,  _enter, _next,  _editor,
 /*  C-p     C-q     C-r     C-s     C-t     C-u     C-v     C-w     */
     _prev,  _quo,   _bsw,   iself,  _mvLw,  killb,  _quo,   _bsw,
 /*  C-x     C-y     C-z     C-[     C-\     C-]     C-^     C-_     */
@@ -139,6 +139,9 @@ inputLineHistSearch(char *prompt, char *def_str, int flag, Hist *hist,
 	CLen = CPos = 0;
     }
 
+#ifdef SUPPORT_WIN9X_CONSOLE_MBCS
+    enable_win9x_console_input();
+#endif
     i_cont = TRUE;
     i_broken = FALSE;
     i_quote = FALSE;
@@ -287,6 +290,10 @@ inputLineHistSearch(char *prompt, char *def_str, int flag, Hist *hist,
     if (need_redraw)
 	displayBuffer(Currentbuf, B_FORCE_REDRAW);
 
+#ifdef SUPPORT_WIN9X_CONSOLE_MBCS
+    disable_win9x_console_input();
+#endif
+
     if (i_broken)
 	return NULL;
 
@@ -302,7 +309,7 @@ inputLineHistSearch(char *prompt, char *def_str, int flag, Hist *hist,
 	    pushHist(hist, p);
     }
     if (flag & IN_FILENAME)
-	return expandName(p);
+	return expandPath(p);
     else
 	return allocStr(p, -1);
 }
@@ -443,15 +450,6 @@ _esc(void)
 	case 'D':
 	    _mvL();
 	    break;
-#if defined(__CYGWIN__) && defined(USE_MOUSE)
-	case 'M':
-	    if ((is_xterm & (NEED_XTERM_ON | NEED_XTERM_OFF)) == NEED_XTERM_ON) {
-		getch();
-		getch();
-		getch();
-	    }
-	    break;
-#endif
 	}
 	break;
     case CTRL_I:
@@ -855,7 +853,7 @@ next_dcompl(int next)
 	    f = Strdup(d);
 	    Strcat_charp(f, CFileBuf[n]);
 	    addstr(conv_from_system(CFileBuf[n]));
-	    if (stat(expandName(f->ptr), &st) != -1 && S_ISDIR(st.st_mode))
+	    if (stat(expandPath(f->ptr), &st) != -1 && S_ISDIR(st.st_mode))
 		addstr("/");
 	}
 	y++;
@@ -959,7 +957,7 @@ doComplete(Str ifn, int *status, int next)
 	if (Strlastchar(CompleteBuf) == '/' && CompleteBuf->length > 1) {
 	    Strshrink(CompleteBuf, 1);
 	}
-	if ((d = opendir(expandName(CompleteBuf->ptr))) == NULL) {
+	if ((d = opendir(expandPath(CompleteBuf->ptr))) == NULL) {
 	    CompleteBuf = Strdup(ifn);
 	    *status = CPL_FAIL;
 	    if (cm_mode & CPL_ON)
@@ -1033,7 +1031,7 @@ doComplete(Str ifn, int *status, int next)
 	    else if (strncmp(p, "file:/", 6) == 0 && p[6] != '/')
 		p = &p[5];
 	}
-	if (stat(expandName(p), &st) != -1 && S_ISDIR(st.st_mode))
+	if (stat(expandPath(p), &st) != -1 && S_ISDIR(st.st_mode))
 	    Strcat_char(CompleteBuf, '/');
     }
     if (cm_mode & CPL_ON)
@@ -1060,6 +1058,8 @@ _prev(void)
 	    return;
 	strCurrentBuf = strBuf;
     }
+    if (DecodeURL)
+	p = url_unquote_conv(p, 0);
     strBuf = Strnew_charp(p);
     CLen = CPos = setStrType(strBuf, strProp);
     offset = 0;
@@ -1077,6 +1077,8 @@ _next(void)
 	return;
     p = nextHist(hist);
     if (p) {
+	if (DecodeURL)
+	    p = url_unquote_conv(p, 0);
 	strBuf = Strnew_charp(p);
     }
     else {
@@ -1125,3 +1127,28 @@ terminated(unsigned char c)
 
     return 0;
 }
+
+static void
+_editor(void)
+{
+    FormItemList fi;
+    char *p;
+
+    if (is_passwd)
+	return;
+
+    fi.readonly = FALSE;
+    fi.value = Strdup(strBuf);
+    Strcat_char(fi.value, '\n');
+
+    input_textarea(&fi);
+
+    strBuf = Strnew();
+    for (p = fi.value->ptr; *p; p++) {
+	if (*p == '\r' || *p == '\n')
+	    continue;
+	Strcat_char(strBuf, *p);
+    }
+    CLen = CPos = setStrType(strBuf, strProp);
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+}
