@@ -1,14 +1,12 @@
-/* $Id: mktable.c,v 1.8.2.1 2002/02/04 16:26:29 ukai Exp $ */
+/* $Id: mktable.c,v 1.11 2002/12/24 17:20:47 ukai Exp $ */
 #include <stdio.h>
 #include <stdlib.h>
-#include <ctype.h>
+#include "myctype.h"
 #include "config.h"
 #include "hash.h"
 #include "Str.h"
 #include <gc.h>
 
-#include "gcmain.c"
-
 /* *INDENT-OFF* */
 defhash(HashItem_ss *, int, hss_i)
 /* *INDENT-ON* */
@@ -27,7 +25,7 @@ defhashfunc(HashItem_ss *, int, hss_i)
 /* *INDENT-ON* */
 
 int
-MAIN(int argc, char *argv[], char **envp)
+main(int argc, char *argv[], char **envp)
 {
     FILE *f;
     Hash_ss *hash;
@@ -76,11 +74,11 @@ MAIN(int argc, char *argv[], char **envp)
 	Strremovetrailingspaces(s);
 	name = Strnew();
 	for (p = s->ptr; *p; p++) {
-	    if (isspace(*p))
+	    if (IS_SPACE(*p))
 		break;
 	    Strcat_char(name, *p);
 	}
-	while (*p && isspace(*p))
+	while (*p && IS_SPACE(*p))
 	    p++;
 	putHash_ss(hash, name->ptr, p);
 	n++;
