@@ -1,4 +1,4 @@
-/* $Id: rc.c,v 1.54 2002/10/28 17:09:29 ukai Exp $ */
+/* $Id: rc.c,v 1.82 2003/02/05 16:48:09 ukai Exp $ */
 /* 
  * Initialization file etc.
  */
@@ -33,8 +33,6 @@ struct rc_search_table {
 static struct rc_search_table *RC_search_table;
 static int RC_table_size;
 
-static char *config_file = NULL;
-
 #define P_INT      0
 #define P_SHORT    1
 #define P_CHARINT  2
@@ -65,7 +63,10 @@ static char *config_file = NULL;
 #define CMT_FRAME        "ե졼μưɽ"
 #define CMT_ARGV_IS_URL  "scheme Τʤ URL Ȥߤʤ"
 #define CMT_TSELF        "target̤ξ_selfѤ"
+#define CMT_OPEN_TAB_BLANK "target_blank_newξϿ֤ǳ"
+#define CMT_OPEN_TAB_DL_LIST "Download list panel 򿷤֤ǳ"
 #define CMT_DISPLINK     "μưɽ"
+#define CMT_DECODE_URL   "URLǥɤɽ"
 #define CMT_DISPLINEINFO "ߤιֹɽ"
 #define CMT_DISP_IMAGE   "饤ɽ"
 #ifdef USE_IMAGE
@@ -74,10 +75,13 @@ static char *config_file = NULL;
 #define CMT_EXT_IMAGE_VIEWER   "ӥ塼ɽ"
 #define CMT_IMAGE_SCALE  "Υ(%)"
 #define CMT_IMGDISPLAY   "ɽ뤿Υޥ"
+#define CMT_IMAGE_MAP_LIST "᡼ޥåפΥɽ"
 #endif
 #define CMT_MULTICOL     "ե̾Υޥɽ"
 #define CMT_ALT_ENTITY   "ƥƥ ASCII ɽɽ"
+#define CMT_GRAPHIC_CHAR "ơ֥˥塼Ȥ graphic ʸȤ"
 #define CMT_FOLD_TEXTAREA "TEXTAREA ιԤޤ֤ɽ"
+#define CMT_DISP_INS_DEL "DEL, S, STRIKE ǤƤɽ롣"
 #define CMT_COLOR        "顼ɽ"
 #define CMT_B_COLOR      "ʸο"
 #define CMT_A_COLOR      "󥫡ο"
@@ -89,7 +93,11 @@ static char *config_file = NULL;
 #define CMT_C_COLOR	 "򤵤Ƥ󥯤ο"
 #define CMT_VISITED_ANCHOR	"ˬ줿Ȥ󥯤ϿѤ"
 #define CMT_V_COLOR	 "ˬ줿Ȥ󥯤ο"
+#define CMT_USE_PROXY    "ץѤ"
 #define CMT_HTTP_PROXY   "HTTPץ(URL)"
+#ifdef USE_SSL
+#define CMT_HTTPS_PROXY  "HTTPSץ(URL)"
+#endif				/* USE_SSL */
 #ifdef USE_GOPHER
 #define CMT_GOPHER_PROXY "GOPHERץ(URL)"
 #endif				/* USE_GOPHER */
@@ -97,11 +105,17 @@ static char *config_file = NULL;
 #define CMT_NO_PROXY     "ץɥᥤ"
 #define CMT_NOPROXY_NETADDR	"ͥåȥɥ쥹ǥץΥå"
 #define CMT_NO_CACHE     "Cache Ȥʤ"
+#ifdef USE_NNTP
+#define CMT_NNTP_SERVER  "News "
+#define CMT_NNTP_MODE    "News ФΥ⡼"
+#define CMT_MAX_NEWS     "News ɽο"
+#endif
 #define CMT_DNS_ORDER	"̾ν"
 #define CMT_DROOT        "/ ɽǥ쥯ȥ(document root)"
 #define CMT_PDROOT       "/~user ɽǥ쥯ȥ"
 #define CMT_CGIBIN       "/cgi-bin ɽǥ쥯ȥ"
 #define CMT_CONFIRM_QQ   "q Ǥνλ˳ǧ"
+#define CMT_CLOSE_TAB_BACK "˥ХåեǸʤ饿֤Ĥ"
 #ifdef USE_MARK
 #define CMT_USE_MARK	"ޡǽͭˤ"
 #endif
@@ -117,6 +131,7 @@ static char *config_file = NULL;
 #ifdef NEXTPAGE_TOPLINE
 #define CMT_NEXTPAGE_TOPLINE	"Υڡ˰ư˥뤬ȥåפˤʤ褦ˤ"
 #endif
+#define CMT_FOLD_LINE    "plain text եιԤޤ֤"
 #define CMT_SHOW_NUM     "ֹɽ"
 #define CMT_SHOW_SRCH_STR "ʸɽ"
 #define CMT_MIMETYPES    "Ѥmime.types"
@@ -127,7 +142,9 @@ static char *config_file = NULL;
 #define CMT_EXTBRZ       "֥饦"
 #define CMT_EXTBRZ2      "֥饦2"
 #define CMT_EXTBRZ3      "֥饦3"
+#define CMT_DISABLE_SECRET_SECURITY_CHECK	"ѥɥեΥѡߥååʤ"
 #define CMT_PASSWDFILE	 "ѥɥե"
+#define CMT_PRE_FORM_FILE	"ʸɹΥեѥե"
 #define CMT_FTPPASS      "FTPΥѥ(̤ϼʬmail addressȤ)"
 #ifdef FTPPASS_HOSTNAMEGEN
 #define CMT_FTPPASS_HOSTNAMEGEN	"FTPΥѥɤΥɥᥤ̾ư"
@@ -157,9 +174,13 @@ static char *config_file = NULL;
 #define CMT_RETRY_HTTP   "URL˼ưŪ http:// 䤦"
 #define CMT_DEFAULT_URL	 "URL򳫤Υǥեʸ"
 #define CMT_DECODE_CTE   "¸ Content-Transfer-Encoding ǥɤ"
+#define CMT_PRESERVE_TIMESTAMP "¸˥ॹפݻ"
 #ifdef USE_MOUSE
 #define CMT_MOUSE         "ޥȤ"
 #define CMT_REVERSE_MOUSE "ޥΥɥåưդˤ"
+#define CMT_RELATIVE_WHEEL_SCROLL "ۥǤΥ®٤ΰ"
+#define CMT_RELATIVE_WHEEL_SCROLL_RATIO "(AΤ)̤β%뤹뤫"
+#define CMT_FIXED_WHEEL_SCROLL_COUNT "(BΤ)뤹Կ"
 #endif				/* USE_MOUSE */
 #define CMT_CLEAR_BUF     "ɽƤʤХåեΥ"
 #define CMT_NOSENDREFERER "Referer: ʤ褦ˤ"
@@ -207,7 +228,10 @@ static char *config_file = NULL;
 #define CMT_FRAME        "Render frames automatically"
 #define CMT_ARGV_IS_URL  "Treat argument without scheme as URL"
 #define CMT_TSELF        "Use _self as default target"
+#define CMT_OPEN_TAB_BLANK "Open link on new tab if target is _blank or _new"
+#define CMT_OPEN_TAB_DL_LIST "Open download list panel on new tab"
 #define CMT_DISPLINK     "Display link URL automatically"
+#define CMT_DECODE_URL   "Display decoded URL"
 #define CMT_DISPLINEINFO "Display current line number"
 #define CMT_DISP_IMAGE   "Display inline images"
 #ifdef USE_IMAGE
@@ -216,10 +240,13 @@ static char *config_file = NULL;
 #define CMT_EXT_IMAGE_VIEWER   "Use external image viewer"
 #define CMT_IMAGE_SCALE  "Scale of image (%)"
 #define CMT_IMGDISPLAY   "External command to display image"
+#define CMT_IMAGE_MAP_LIST "Use link list of image map"
 #endif
 #define CMT_MULTICOL     "Display file names in multi-column format"
 #define CMT_ALT_ENTITY   "Use ASCII equivalents to display entities"
+#define CMT_GRAPHIC_CHAR "Use graphic char for border of table and menu"
 #define CMT_FOLD_TEXTAREA "Fold lines in TEXTAREA"
+#define CMT_DISP_INS_DEL "Display DEL, S and STRIKE element"
 #define CMT_COLOR        "Display with color"
 #define CMT_B_COLOR      "Color of normal character"
 #define CMT_A_COLOR      "Color of anchor"
@@ -231,7 +258,11 @@ static char *config_file = NULL;
 #define CMT_V_COLOR	 "Color of visited link"
 #define CMT_BG_COLOR     "Color of background"
 #define CMT_MARK_COLOR   "Color of mark"
+#define CMT_USE_PROXY    "Use proxy"
 #define CMT_HTTP_PROXY   "URL of HTTP proxy host"
+#ifdef USE_SSL
+#define CMT_HTTPS_PROXY  "URL of HTTPS proxy host"
+#endif				/* USE_SSL */
 #ifdef USE_GOPHER
 #define CMT_GOPHER_PROXY "URL of GOPHER proxy host"
 #endif				/* USE_GOPHER */
@@ -239,11 +270,17 @@ static char *config_file = NULL;
 #define CMT_NO_PROXY     "Domains to be accessed directly (no proxy)"
 #define CMT_NOPROXY_NETADDR	"Check noproxy by network address"
 #define CMT_NO_CACHE     "Disable cache"
+#ifdef USE_NNTP
+#define CMT_NNTP_SERVER  "News server"
+#define CMT_NNTP_MODE    "Mode of news server"
+#define CMT_MAX_NEWS     "Number of news messages"
+#endif
 #define CMT_DNS_ORDER	"Order of name resolution"
 #define CMT_DROOT        "Directory corresponding to / (document root)"
 #define CMT_PDROOT       "Directory corresponding to /~user"
 #define CMT_CGIBIN       "Directory corresponding to /cgi-bin"
 #define CMT_CONFIRM_QQ   "Confirm when quitting with q"
+#define CMT_CLOSE_TAB_BACK "Close tab if buffer is last when back"
 #ifdef USE_MARK
 #define CMT_USE_MARK	"Enable mark operations"
 #endif
@@ -259,6 +296,7 @@ static char *config_file = NULL;
 #ifdef NEXTPAGE_TOPLINE
 #define CMT_NEXTPAGE_TOPLINE	"Move cursor to top line when moving to next page"
 #endif
+#define CMT_FOLD_LINE    "Fold lines of plain text file"
 #define CMT_SHOW_NUM     "Show line numbers"
 #define CMT_SHOW_SRCH_STR "Show search string"
 #define CMT_MIMETYPES    "List of mime.types files"
@@ -269,7 +307,9 @@ static char *config_file = NULL;
 #define CMT_EXTBRZ       "External Browser"
 #define CMT_EXTBRZ2      "Second External Browser"
 #define CMT_EXTBRZ3      "Third External Browser"
+#define CMT_DISABLE_SECRET_SECURITY_CHECK	"Disable secret file security check"
 #define CMT_PASSWDFILE	 "Password file"
+#define CMT_PRE_FORM_FILE	"File for setting form on loading"
 #define CMT_FTPPASS      "Password for anonymous FTP (your mail address)"
 #ifdef FTPPASS_HOSTNAMEGEN
 #define CMT_FTPPASS_HOSTNAMEGEN "Generate domain part of password for FTP"
@@ -299,9 +339,13 @@ static char *config_file = NULL;
 #define CMT_RETRY_HTTP   "Prepend http:// to URL automatically"
 #define CMT_DEFAULT_URL  "Default value for open-URL command"
 #define CMT_DECODE_CTE   "Decode Content-Transfer-Encoding when saving"
+#define CMT_PRESERVE_TIMESTAMP "Preserve timestamp when saving"
 #ifdef USE_MOUSE
 #define CMT_MOUSE         "Enable mouse"
 #define CMT_REVERSE_MOUSE "Scroll in reverse direction of mouse drag"
+#define CMT_RELATIVE_WHEEL_SCROLL "Behavior of wheel scroll speed"
+#define CMT_RELATIVE_WHEEL_SCROLL_RATIO "(A only)Scroll by # % of screen"
+#define CMT_FIXED_WHEEL_SCROLL_COUNT "(B only)Scroll by # lines"
 #endif				/* USE_MOUSE */
 #define CMT_CLEAR_BUF     "Free memory of undisplayed buffers"
 #define CMT_NOSENDREFERER "Suppress `Referer:' header"
@@ -425,6 +469,20 @@ static struct sel_c defaulturls[] = {
 #endif
     {0, NULL, NULL}
 };
+#ifdef USE_MOUSE
+static struct sel_c wheelmode[] = {
+#if LANG == JA
+    {TRUE, "1", "A:̥"},
+    {FALSE, "0", "B:ιԿ"},
+    {0, NULL, NULL}
+#else				/* LANG != JA */
+    {TRUE, "1", "A:relative to screen height"},
+    {FALSE, "0", "B:fixed speed"},
+    {0, NULL, NULL}
+#endif				/* LANG != JA */
+};
+#endif				/* MOUSE */
+
 #ifdef INET6
 static struct sel_c dnsorders[] = {
     {N_S(DNS_ORDER_UNSPEC), "unspecified"},
@@ -463,8 +521,13 @@ struct param_ptr params1[] = {
 #endif				/* JP_CHARSET */
     {"frame", P_CHARINT, PI_ONOFF, (void *)&RenderFrame, CMT_FRAME, NULL},
     {"target_self", P_CHARINT, PI_ONOFF, (void *)&TargetSelf, CMT_TSELF, NULL},
+    {"open_tab_blank", P_INT, PI_ONOFF, (void *)&open_tab_blank,
+     CMT_OPEN_TAB_BLANK, NULL},
+    {"open_tab_dl_list", P_INT, PI_ONOFF, (void *)&open_tab_dl_list,
+     CMT_OPEN_TAB_DL_LIST, NULL},
     {"display_link", P_INT, PI_ONOFF, (void *)&displayLink, CMT_DISPLINK,
      NULL},
+    {"decode_url", P_INT, PI_ONOFF, (void *)&DecodeURL, CMT_DECODE_URL, NULL},
     {"display_lineinfo", P_INT, PI_ONOFF, (void *)&displayLineInfo,
      CMT_DISPLINEINFO, NULL},
     {"ext_dirlist", P_INT, PI_ONOFF, (void *)&UseExternalDirBuffer,
@@ -480,8 +543,14 @@ struct param_ptr params1[] = {
     {"multicol", P_INT, PI_ONOFF, (void *)&multicolList, CMT_MULTICOL, NULL},
     {"alt_entity", P_CHARINT, PI_ONOFF, (void *)&UseAltEntity, CMT_ALT_ENTITY,
      NULL},
+#ifndef KANJI_SYMBOLS
+    {"graphic_char", P_CHARINT, PI_ONOFF, (void *)&UseGraphicChar,
+     CMT_GRAPHIC_CHAR, NULL},
+#endif
     {"fold_textarea", P_CHARINT, PI_ONOFF, (void *)&FoldTextarea,
      CMT_FOLD_TEXTAREA, NULL},
+    {"display_ins_del", P_INT, PI_ONOFF, (void *)&displayInsDel,
+     CMT_DISP_INS_DEL, NULL},
     {"ignore_null_img_alt", P_INT, PI_ONOFF, (void *)&ignore_null_img_alt,
      CMT_IGNORE_NULL_IMG_ALT, NULL},
     {"view_unseenobject", P_INT, PI_ONOFF, (void *)&view_unseenobject,
@@ -499,7 +568,10 @@ struct param_ptr params1[] = {
      NULL},
     {"imgdisplay", P_STRING, PI_TEXT, (void *)&Imgdisplay, CMT_IMGDISPLAY,
      NULL},
+    {"image_map_list", P_INT, PI_ONOFF, (void *)&image_map_list,
+     CMT_IMAGE_MAP_LIST, NULL},
 #endif
+    {"fold_line", P_INT, PI_ONOFF, (void *)&FoldLine, CMT_FOLD_LINE, NULL},
     {"show_lnum", P_INT, PI_ONOFF, (void *)&showLineNum, CMT_SHOW_NUM, NULL},
     {"show_srch_str", P_INT, PI_ONOFF, (void *)&show_srch_str,
      CMT_SHOW_SRCH_STR, NULL},
@@ -551,6 +623,8 @@ struct param_ptr params3[] = {
 #endif				/* USE_HISTORY */
     {"confirm_qq", P_INT, PI_ONOFF, (void *)&confirm_on_quit, CMT_CONFIRM_QQ,
      NULL},
+    {"close_tab_back", P_INT, PI_ONOFF, (void *)&close_tab_back,
+     CMT_CLOSE_TAB_BACK, NULL},
 #ifdef USE_MARK
     {"mark", P_INT, PI_ONOFF, (void *)&use_mark, CMT_USE_MARK, NULL},
 #endif
@@ -577,19 +651,35 @@ struct param_ptr params3[] = {
     {"use_mouse", P_INT, PI_ONOFF, (void *)&use_mouse, CMT_MOUSE, NULL},
     {"reverse_mouse", P_INT, PI_ONOFF, (void *)&reverse_mouse,
      CMT_REVERSE_MOUSE, NULL},
+    {"relative_wheel_scroll", P_INT, PI_SEL_C, (void *)&relative_wheel_scroll,
+     CMT_RELATIVE_WHEEL_SCROLL, wheelmode},
+    {"relative_wheel_scroll_ratio", P_INT, PI_TEXT,
+     (void *)&relative_wheel_scroll_ratio,
+     CMT_RELATIVE_WHEEL_SCROLL_RATIO, NULL},
+    {"fixed_wheel_scroll_count", P_INT, PI_TEXT,
+     (void *)&fixed_wheel_scroll_count,
+     CMT_FIXED_WHEEL_SCROLL_COUNT, NULL},
 #endif				/* USE_MOUSE */
     {"clear_buffer", P_INT, PI_ONOFF, (void *)&clear_buffer, CMT_CLEAR_BUF,
      NULL},
     {"decode_cte", P_CHARINT, PI_ONOFF, (void *)&DecodeCTE, CMT_DECODE_CTE,
      NULL},
+    {"preserve_timestamp", P_CHARINT, PI_ONOFF, (void *)&PreserveTimestamp,
+     CMT_PRESERVE_TIMESTAMP, NULL},
     {"keymap_file", P_STRING, PI_TEXT, (void *)&keymap_file, CMT_KEYMAP_FILE,
      NULL},
     {NULL, 0, 0, NULL, NULL, NULL},
 };
 
 struct param_ptr params4[] = {
+    {"use_proxy", P_CHARINT, PI_ONOFF, (void *)&use_proxy, CMT_USE_PROXY,
+     NULL},
     {"http_proxy", P_STRING, PI_TEXT, (void *)&HTTP_proxy, CMT_HTTP_PROXY,
      NULL},
+#ifdef USE_SSL
+    {"https_proxy", P_STRING, PI_TEXT, (void *)&HTTPS_proxy, CMT_HTTPS_PROXY,
+     NULL},
+#endif				/* USE_SSL */
 #ifdef USE_GOPHER
     {"gopher_proxy", P_STRING, PI_TEXT, (void *)&GOPHER_proxy,
      CMT_GOPHER_PROXY, NULL},
@@ -672,11 +762,16 @@ struct param_ptr params8[] = {
 struct param_ptr params9[] = {
     {"passwd_file", P_STRING, PI_TEXT, (void *)&passwd_file, CMT_PASSWDFILE,
      NULL},
+    {"disable_secret_security_check", P_INT, PI_ONOFF,
+     (void *)&disable_secret_security_check, CMT_DISABLE_SECRET_SECURITY_CHECK,
+     NULL},
     {"ftppasswd", P_STRING, PI_TEXT, (void *)&ftppasswd, CMT_FTPPASS, NULL},
 #ifdef FTPPASS_HOSTNAMEGEN
     {"ftppass_hostnamegen", P_INT, PI_ONOFF, (void *)&ftppass_hostnamegen,
      CMT_FTPPASS_HOSTNAMEGEN, NULL},
 #endif
+    {"pre_form_file", P_STRING, PI_TEXT, (void *)&pre_form_file,
+     CMT_PRE_FORM_FILE, NULL},
     {"user_agent", P_STRING, PI_TEXT, (void *)&UserAgent, CMT_USERAGENT, NULL},
     {"no_referer", P_INT, PI_ONOFF, (void *)&NoSendReferer, CMT_NOSENDREFERER,
      NULL},
@@ -702,6 +797,12 @@ struct param_ptr params9[] = {
     {"dns_order", P_INT, PI_SEL_C, (void *)&DNS_order, CMT_DNS_ORDER,
      dnsorders},
 #endif				/* INET6 */
+#ifdef USE_NNTP
+    {"nntpserver", P_STRING, PI_TEXT, (void *)&NNTP_server, CMT_NNTP_SERVER,
+     NULL},
+    {"nntpmode", P_STRING, PI_TEXT, (void *)&NNTP_mode, CMT_NNTP_MODE, NULL},
+    {"max_news", P_INT, PI_TEXT, (void *)&MaxNewsMessage, CMT_MAX_NEWS, NULL},
+#endif
     {NULL, 0, 0, NULL, NULL, NULL},
 };
 
@@ -750,7 +851,7 @@ compare_table(struct rc_search_table *a, struct rc_search_table *b)
     return strcmp(a->param->name, b->param->name);
 }
 
-void
+static void
 create_option_search_table()
 {
     int i, j, k;
@@ -910,18 +1011,18 @@ str_to_bool(char *value, int old)
 {
     if (value == NULL)
 	return 1;
-    switch (tolower(*value)) {
+    switch (TOLOWER(*value)) {
     case '0':
     case 'f':			/* false */
     case 'n':			/* no */
     case 'u':			/* undef */
 	return 0;
     case 'o':
-	if (tolower(value[1]) == 'f')	/* off */
+	if (TOLOWER(value[1]) == 'f')	/* off */
 	    return 0;
 	return 1;		/* on */
     case 't':
-	if (tolower(value[1]) == 'o')	/* toggle */
+	if (TOLOWER(value[1]) == 'o')	/* toggle */
 	    return !old;
 	return 1;		/* true */
     case '!':
@@ -938,7 +1039,7 @@ str_to_color(char *value)
 {
     if (value == NULL)
 	return 8;		/* terminal */
-    switch (tolower(*value)) {
+    switch (TOLOWER(*value)) {
     case '0':
 	return 0;		/* black */
     case '1':
@@ -1175,6 +1276,10 @@ parse_proxy()
 {
     if (non_null(HTTP_proxy))
 	parseURL(HTTP_proxy, &HTTP_proxy_parsed, NULL);
+#ifdef USE_SSL
+    if (non_null(HTTPS_proxy))
+	parseURL(HTTPS_proxy, &HTTPS_proxy_parsed, NULL);
+#endif				/* USE_SSL */
 #ifdef USE_GOPHER
     if (non_null(GOPHER_proxy))
 	parseURL(GOPHER_proxy, &GOPHER_proxy_parsed, NULL);
@@ -1220,6 +1325,8 @@ do_mkdir(const char *dir, long mode)
 void
 sync_with_option(void)
 {
+    if (PagerMax < LINES)
+	PagerMax = LINES;
     WrapSearch = WrapDefault;
     parse_proxy();
 #ifdef USE_COOKIE
@@ -1240,6 +1347,7 @@ sync_with_option(void)
     displayImage = FALSE;	/* XXX */
 #endif
     loadPasswd();
+    loadPreForm();
 
     if (AcceptLang == NULL || *AcceptLang == '\0') {
 #if LANG == JA
@@ -1252,28 +1360,37 @@ sync_with_option(void)
 	AcceptEncoding = acceptableEncoding();
     if (AcceptMedia == NULL || *AcceptMedia == '\0')
 	AcceptMedia = acceptableMimeTypes();
-    initKeymap(FALSE);
+    if (fmInitialized) {
+	initKeymap(FALSE);
+#ifdef USE_MOUSE
+	initMouseAction();
+#endif				/* MOUSE */
+#ifdef USE_MENU
+	initMenu();
+#endif				/* MENU */
+    }
 }
 
 void
-init_rc(char *config_filename)
+init_rc(void)
 {
+    int i;
     struct stat st;
     FILE *f;
-    char *tmpdir;
 
-    if (((tmpdir = getenv("TMP")) == NULL || *tmpdir == '\0')
-	&& ((tmpdir = getenv("TEMP")) == NULL || *tmpdir == '\0')
-	&& ((tmpdir = getenv("TMPDIR")) == NULL || *tmpdir == '\0'))
-	tmpdir = "/tmp";
+    if (rc_dir != NULL)
+	goto open_rc;
+
+    rc_dir = expandPath(RC_DIR);
+    i = strlen(rc_dir);
+    if (i > 1 && rc_dir[i - 1] == '/')
+	rc_dir[i - 1] = '\0';
 
     if (stat(rc_dir, &st) < 0) {
 	if (errno == ENOENT) {	/* no directory */
 	    if (do_mkdir(rc_dir, 0700) < 0) {
 		fprintf(stderr, "Can't create config directory (%s)!", rc_dir);
-		rc_dir = tmpdir;
-		rc_dir_is_tmp = TRUE;
-		return;
+		goto rc_dir_err;
 	    }
 	    else {
 		stat(rc_dir, &st);
@@ -1281,38 +1398,56 @@ init_rc(char *config_filename)
 	}
 	else {
 	    fprintf(stderr, "Can't open config directory (%s)!", rc_dir);
-	    rc_dir = tmpdir;
-	    rc_dir_is_tmp = TRUE;
-	    return;
+	    goto rc_dir_err;
 	}
     }
     if (!S_ISDIR(st.st_mode)) {
 	/* not a directory */
 	fprintf(stderr, "%s is not a directory!", rc_dir);
-	rc_dir = tmpdir;
-	rc_dir_is_tmp = TRUE;
-	return;
+	goto rc_dir_err;
     }
+    if (!(st.st_mode & S_IWUSR)) {
+	fprintf(stderr, "%s is not writable!", rc_dir);
+	goto rc_dir_err;
+    }
+    no_rc_dir = FALSE;
+    tmp_dir = rc_dir;
 
+    if (config_file == NULL)
+	config_file = rcFile(CONFIG_FILE);
+
+    create_option_search_table();
+
+  open_rc:
     /* open config file */
     if ((f = fopen(etcFile(W3MCONFIG), "rt")) != NULL) {
 	interpret_rc(f);
 	fclose(f);
     }
-    config_file = config_filename;
-    if (config_file == NULL)
-	config_file = rcFile(CONFIG_FILE);
     if ((f = fopen(config_file, "rt")) != NULL) {
 	interpret_rc(f);
 	fclose(f);
     }
-    sync_with_option();
+    return;
+
+  rc_dir_err:
+    no_rc_dir = TRUE;
+    if (((tmp_dir = getenv("TMPDIR")) == NULL || *tmp_dir == '\0') &&
+	((tmp_dir = getenv("TMP")) == NULL || *tmp_dir == '\0') &&
+	((tmp_dir = getenv("TEMP")) == NULL || *tmp_dir == '\0'))
+	tmp_dir = "/tmp";
 }
 
 
 static char optionpanel_src1[] =
-    "<html><head><title>Option Setting Panel</title></head>\
-<body><center><b>Option Setting Panel</b><br><b>(w3m version %s)</b></center><p>\n" "<a href=\"file:///$LIB/" W3MHELPERPANEL_CMDNAME "?mode=panel\">%s</a>\n" "<form method=internal action=option>";
+    "<html><head><title>Option Setting Panel</title></head><body>\
+<h1 align=center>Option Setting Panel<br>(w3m version %s)</b></h1>\
+<form method=post action=\"file:///$LIB/" W3MHELPERPANEL_CMDNAME "\">\
+<input type=hidden name=mode value=panel>\
+<input type=hidden name=cookie value=\"%s\">\
+<input type=submit value=\"%s\">\
+</form><br>\
+<form method=internal action=option>";
 
 static Str
 to_str(struct param_ptr *p)
@@ -1349,7 +1484,8 @@ to_str(struct param_ptr *p)
 Buffer *
 load_option_panel(void)
 {
-    Str src = Sprintf(optionpanel_src1, w3m_version, CMT_HELPER);
+    Str src = Sprintf(optionpanel_src1, html_quote(w3m_version),
+		      html_quote(Local_cookie->ptr), CMT_HELPER);
     struct param_ptr *p;
     struct sel_c *s;
     int x, i;
@@ -1420,7 +1556,7 @@ panel_set_option(struct parsed_tagarg *arg)
 {
     FILE *f = NULL;
 
-    if (rc_dir_is_tmp) {
+    if (no_rc_dir) {
 	disp_message("There's no ~/.w3m directory... config not saved", FALSE);
     }
     else {
@@ -1450,40 +1586,35 @@ rcFile(char *base)
 	 (base[0] == '.'
 	  && (base[1] == '/' || (base[1] == '.' && base[2] == '/')))
 	 || (base[0] == '~' && base[1] == '/')))
-	return expandName(base);
-    else {
-	Str file = Strnew_charp(rc_dir);
-
-	if (Strlastchar(file) != '/')
-	    Strcat_char(file, '/');
-	Strcat_charp(file, base);
-	return expandName(file->ptr);
-    }
+	/* /file, ./file, ../file, ~/file */
+	return expandPath(base);
+    return expandPath(Strnew_m_charp(rc_dir, "/", base, NULL)->ptr);
 }
 
+char *
+auxbinFile(char *base)
+{
+    return expandPath(Strnew_m_charp(w3m_auxbin_dir(), "/", base, NULL)->ptr);
+}
+
+#if 0				/* not used */
 char *
 libFile(char *base)
 {
-    Str file = Strnew_charp(w3m_lib_dir());
-    Strcat_char(file, '/');
-    Strcat_charp(file, base);
-    return expandName(file->ptr);
+    return expandPath(Strnew_m_charp(w3m_lib_dir(), "/", base, NULL)->ptr);
 }
+#endif
 
 char *
 etcFile(char *base)
 {
-    Str file = Strnew_charp(w3m_etc_dir());
-    Strcat_char(file, '/');
-    Strcat_charp(file, base);
-    return expandName(file->ptr);
+    return expandPath(Strnew_m_charp(w3m_etc_dir(), "/", base, NULL)->ptr);
 }
 
+#ifndef USE_HELP_CGI
 char *
 helpFile(char *base)
 {
-    Str file = Strnew_charp(w3m_help_dir());
-    Strcat_char(file, '/');
-    Strcat_charp(file, base);
-    return expandName(file->ptr);
+    return expandPath(Strnew_m_charp(w3m_help_dir(), "/", base, NULL)->ptr);
 }
+#endif
