@@ -1,4 +1,4 @@
-/* $Id: terms.c,v 1.33 2002/09/09 13:49:17 ukai Exp $ */
+/* $Id: terms.c,v 1.48 2003/02/23 16:00:15 ukai Exp $ */
 /* 
  * An original curses library for EUC-kanji by Akinori ITO,     December 1989
  * revised by Akinori ITO, January 1995
@@ -33,10 +33,7 @@ static int cwidth = 8, cheight = 16;
 static int xpix, ypix, nbs, obs = 0;
 #endif				/* use_SYSMOUSE */
 
-#ifndef __CYGWIN__
-static
-#endif
-int is_xterm = 0;
+static int is_xterm = 0;
 
 void mouse_init(), mouse_end();
 int mouseActive = 0;
@@ -61,35 +58,24 @@ extern int CodePage;
 
 #if defined(__CYGWIN__)
 #include <windows.h>
-static HANDLE hConIn;
-static int isWin95;
-static int isWinConsole;
-static char *ConInV;
-static int iConIn, nConIn, nConInMax;
+#include <sys/cygwin.h>
+static int isWinConsole = 0;
+static int isLocalConsole = 0;
 #ifdef USE_MOUSE
-static MOUSE_EVENT_RECORD lastConMouse;
+int cygwin_mouse_btn_swapped = 0;
 #endif
 
-void
-check_win32_console(void)
-{
-    char *tty;
+#if defined(SUPPORT_WIN9X_CONSOLE_MBCS)
+static HANDLE hConIn = INVALID_HANDLE_VALUE;
+static int isWin95 = 0;
+static char *ConInV;
+static int iConIn, nConIn, nConInMax;
 
-    tty = ttyname(1);
-    if (!strncmp(tty, "/dev/con", 8)) {
-	isWinConsole = TRUE;
-    }
-    else {
-	isWinConsole = FALSE;
-    }
-}
-
-void
-init_win32_console_handle(void)
+static void
+check_win9x(void)
 {
     OSVERSIONINFO winVersionInfo;
 
-    check_win32_console();
     winVersionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
     if (GetVersionEx(&winVersionInfo) == 0) {
 	fprintf(stderr, "can't get Windows version information.\n");
@@ -97,20 +83,36 @@ init_win32_console_handle(void)
     }
     if (winVersionInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) {
 	isWin95 = 1;
-    }
-    hConIn = INVALID_HANDLE_VALUE;
-    if (isWin95) {
-	if (isWinConsole) {
-	    if (isatty(0)) {
-		hConIn = GetStdHandle(STD_INPUT_HANDLE);
-	    }
-	    else {
-		hConIn = CreateFile("CONIN$", GENERIC_READ,
-				    FILE_SHARE_READ,
-				    NULL, OPEN_EXISTING, 0, NULL);
-	    }
+	if (ttyslot() != -1) {
+	    isLocalConsole = 0;
 	}
     }
+    else {
+	isWin95 = 0;
+    }
+}
+
+void
+enable_win9x_console_input(void)
+{
+    if (isWin95 && isWinConsole && isLocalConsole &&
+	hConIn == INVALID_HANDLE_VALUE) {
+	hConIn = CreateFile("CONIN$", GENERIC_READ | GENERIC_WRITE,
+			    FILE_SHARE_READ | FILE_SHARE_WRITE,
+			    NULL, OPEN_EXISTING, 0, NULL);
+	if (hConIn != INVALID_HANDLE_VALUE) {
+	    getch();
+	}
+    }
+}
+
+void
+disable_win9x_console_input(void)
+{
+    if (hConIn != INVALID_HANDLE_VALUE) {
+	CloseHandle(hConIn);
+	hConIn = INVALID_HANDLE_VALUE;
+    }
 }
 
 static void
@@ -143,44 +145,6 @@ read_win32_console_input(void)
 	    }
 
 	    break;
-#ifdef USE_MOUSE
-	case MOUSE_EVENT:
-	    if ((lastConMouse.dwButtonState ^ rec.Event.MouseEvent.
-		 dwButtonState) & ~(~0 << 5)) {
-		int down;
-		MOUSE_EVENT_RECORD *mer;
-		INPUT_RECORD dummy;
-
-		expand_win32_console_input_buffer(6);
-		mer = &rec.Event.MouseEvent;
-		ConInV[nConIn] = '\033';
-		ConInV[nConIn + 1] = '[';
-		ConInV[nConIn + 2] = 'M';
-
-		if (~(mer->dwButtonState) & lastConMouse.
-		    dwButtonState & ~(~0 << 5))
-		    ConInV[nConIn + 3] = MOUSE_BTN_UP + ' ';
-		else if (!
-			 (down =
-			  mer->dwButtonState & ~lastConMouse.
-			  dwButtonState & ~(~0 << 5)))
-		    break;
-		else
-		    ConInV[nConIn + 3] = (down & (1 << 0) ? MOUSE_BTN1_DOWN :
-					  down & (1 << 1) ? MOUSE_BTN3_DOWN :
-					  down & (1 << 2) ? MOUSE_BTN2_DOWN :
-					  down & (1 << 3) ?
-					  MOUSE_BTN4_DOWN_XTERM :
-					  MOUSE_BTN5_DOWN_XTERM) + ' ';
-
-		ConInV[nConIn + 4] = mer->dwMousePosition.X + '!';
-		ConInV[nConIn + 5] = mer->dwMousePosition.Y + '!';
-		nConIn += 6;
-		lastConMouse = *mer;
-		ReadConsoleInput(hConIn, &rec, 1, &nevents);
-		return 6;
-	    }
-#endif
 	default:
 	    break;
 	}
@@ -190,7 +154,7 @@ read_win32_console_input(void)
     return 0;
 }
 
-int
+static int
 read_win32_console(char *s, int n)
 {
     KEY_EVENT_RECORD *ker;
@@ -220,93 +184,69 @@ read_win32_console(char *s, int n)
     return n;
 }
 
-static int
-cmp_tv(const struct timeval *tv_a, const struct timeval *tv_b)
+#endif				/* SUPPORT_WIN9X_CONSOLE_MBCS */
+
+static HWND
+GetConsoleHwnd(void)
 {
-    return ((tv_a->tv_sec < tv_b->tv_sec) ? -1 :
-	    (tv_a->tv_sec > tv_b->tv_sec) ? 1 :
-	    (tv_a->tv_usec < tv_b->tv_usec) ? -1 :
-	    (tv_a->tv_usec > tv_b->tv_usec) ? 1 : 0);
+#define MY_BUFSIZE 1024
+    HWND hwndFound;
+    char pszNewWindowTitle[MY_BUFSIZE];
+    char pszOldWindowTitle[MY_BUFSIZE];
+
+    GetConsoleTitle(pszOldWindowTitle, MY_BUFSIZE);
+    wsprintf(pszNewWindowTitle, "%d/%d",
+	     GetTickCount(), GetCurrentProcessId());
+    SetConsoleTitle(pszNewWindowTitle);
+    Sleep(40);
+    hwndFound = FindWindow(NULL, pszNewWindowTitle);
+    SetConsoleTitle(pszOldWindowTitle);
+    return (hwndFound);
 }
 
-static int
-subtract_tv(struct timeval *dst, const struct timeval *src)
+#ifdef USE_MOUSE
+static unsigned long
+cygwin_version(void)
 {
-    if ((dst->tv_usec -= src->tv_usec) < 0) {
-	--(dst->tv_sec);
-	dst->tv_usec += 1000000;
+    struct per_process *p;
+
+    p = (struct per_process *)cygwin_internal(CW_USER_DATA);
+    if (p != NULL) {
+	return (p->dll_major * 1000) + p->dll_minor;
     }
-
-    return ((dst->tv_sec -= src->tv_sec) < 0 ? -1 :
-	    !dst->tv_sec ? (dst->tv_usec <
-			    0 ? -1 : !dst->tv_usec ? 0 : 1) : 1);
-}
-
-int
-select_or_poll_win32_console(int n, fd_set * rfds, fd_set * wfds,
-			     fd_set * efds, struct timeval *tout)
-{
-    int m;
-    DWORD nevents;
-
-    if ((m = select(n, rfds, wfds, efds, tout)) < 0)
-	return m;
-
-    if (iConIn < nConIn) {
-	FD_SET(tty, rfds);
-	++m;
-    }
-    else {
-	iConIn = nConIn = 0;
-
-	while (GetNumberOfConsoleInputEvents(hConIn, &nevents) && nevents)
-	    read_win32_console_input();
-
-	if (nConIn) {
-	    FD_SET(tty, rfds);
-	    ++m;
-	}
-    }
-
-    return m;
-}
-
-int
-select_win32_console(int n, fd_set * rfds, fd_set * wfds, fd_set * efds,
-		     struct timeval *tout)
-{
-    static struct timeval polltv = { 0, 1000000 / CLOCKS_PER_SEC };
-    int m;
-    struct timeval tv;
-
-    if (hConIn == INVALID_HANDLE_VALUE || tty < 0 || tty >= n || !rfds
-	|| !FD_ISSET(tty, rfds))
-	return select(n, rfds, wfds, efds, tout);
-
-    FD_CLR(tty, rfds);
-
-    if (tout) {
-	while (cmp_tv(tout, &polltv) > 0) {
-	    tv = polltv;
-
-	    if ((m = select_or_poll_win32_console(n, rfds, wfds, efds, &tv)))
-		return m;
-
-	    subtract_tv(tout, &polltv);
-	}
-
-	return select_or_poll_win32_console(n, rfds, wfds, efds, tout);
-    }
-    else
-	for (;;) {
-	    tv = polltv;
-
-	    if ((m = select_or_poll_win32_console(n, rfds, wfds, efds, &tv)))
-		return m;
-	}
+    return 0;
 }
 #endif
 
+static void
+check_cygwin_console(void)
+{
+    char *term = getenv("TERM");
+    HANDLE hWnd;
+
+    if (strncmp(term, "cygwin", 6) == 0) {
+	isWinConsole = 1;
+    }
+    if (isWinConsole) {
+	hWnd = GetConsoleHwnd();
+	if (hWnd != INVALID_HANDLE_VALUE) {
+	    if (IsWindowVisible(hWnd)) {
+		isLocalConsole = 1;
+	    }
+	}
+#ifdef SUPPORT_WIN9X_CONSOLE_MBCS
+	check_win9x();
+#endif
+    }
+#ifdef USE_MOUSE
+    if (cygwin_version() <= 1003015) {
+	/* cygwin DLL 1.3.15 or earler */
+	cygwin_mouse_btn_swapped = 1;
+    }
+#endif
+}
+#endif				/* __CYGWIN__ */
+
 char *getenv(const char *);
 MySignalHandler reset_exit(SIGNAL_ARG), error_dump(SIGNAL_ARG);
 void setlinescols(void);
@@ -422,12 +362,9 @@ typedef struct scline {
     short eol;
 } Screen;
 
-static
-TerminalMode d_ioval;
-static
-int tty;
-static
-FILE *ttyf;
+static TerminalMode d_ioval;
+static int tty = -1;
+static FILE *ttyf = NULL;
 
 static
 char bp[1024], funcstr[256];
@@ -461,7 +398,10 @@ void clear(), wrap(), touch_line(), touch_column(int);
 void switch_wchar(FILE *);
 void switch_ascii(FILE *);
 #endif
-void need_clrtoeol(void), clrtoeol(void);	/* conflicts with curs_clear(3)? */
+#if 0
+void need_clrtoeol(void);
+#endif
+void clrtoeol(void);		/* conflicts with curs_clear(3)? */
 
 int write1(char);
 
@@ -477,12 +417,21 @@ writestr(char *s)
 
 #ifdef USE_MOUSE
 #define W3M_TERM_INFO(name, title, mouse)	name, title, mouse
+#define NEED_XTERM_ON   (1)
+#define NEED_XTERM_OFF  (1<<1)
+#ifdef __CYGWIN__
+#define NEED_CYGWIN_ON  (1<<2)
+#define NEED_CYGWIN_OFF (1<<3)
+#endif
 #else
 #define W3M_TERM_INFO(name, title, mouse)	name, title
 #endif
 
-#define XTERM_TITLE	"\033]0;w3m: %s\007"
-#define SCREEN_TITLE	"\033k%s\033\134"
+static char XTERM_TITLE[] = "\033]0;w3m: %s\007";
+static char SCREEN_TITLE[] = "\033k%s\033\134";
+#ifdef __CYGWIN__
+static char CYGWIN_TITLE[] = "w3m: %s";
+#endif
 
 /* *INDENT-OFF* */
 static struct w3m_term_info {
@@ -499,7 +448,7 @@ static struct w3m_term_info {
     {W3M_TERM_INFO("mlterm", XTERM_TITLE, (NEED_XTERM_ON|NEED_XTERM_OFF))},
     {W3M_TERM_INFO("screen", SCREEN_TITLE, 0)},
 #ifdef __CYGWIN__
-    {W3M_TERM_INFO("cygwin", NULL, NEED_XTERM_ON)},
+    {W3M_TERM_INFO("cygwin", CYGWIN_TITLE, (NEED_CYGWIN_ON|NEED_CYGWIN_OFF))},
 #endif
     {W3M_TERM_INFO(NULL, NULL, 0)}
 };
@@ -522,7 +471,7 @@ set_tty(void)
     }
     ttyf = fdopen(tty, "w");
 #ifdef __CYGWIN__
-    init_win32_console_handle();
+    check_cygwin_console();
 #endif
     TerminalGet(tty, &d_ioval);
     if (displayTitleTerm != NULL) {
@@ -607,7 +556,14 @@ set_cc(int spec, int val)
 void
 close_tty(void)
 {
-    close(tty);
+    if (tty > 2)
+	close(tty);
+}
+
+char *
+ttyname_tty(void)
+{
+    return ttyname(tty);
 }
 
 void
@@ -627,9 +583,9 @@ reset_tty(void)
 	    writestr(T_cl);
     }
     writestr(T_se);		/* reset terminal */
-    fflush(ttyf);
+    flush_tty();
     TerminalSet(tty, &d_ioval);
-    close(tty);
+    close_tty();
 }
 
 MySignalHandler
@@ -647,7 +603,7 @@ reset_exit(SIGNAL_ARG)
 MySignalHandler
 error_dump(SIGNAL_ARG)
 {
-    signal(SIGIOT, SIG_DFL);
+    mySignal(SIGIOT, SIG_DFL);
     reset_tty();
     abort();
     SIGNAL_RETURN;
@@ -656,17 +612,17 @@ error_dump(SIGNAL_ARG)
 void
 set_int(void)
 {
-    signal(SIGHUP, reset_exit);
-    signal(SIGINT, reset_exit);
-    signal(SIGQUIT, reset_exit);
-    signal(SIGTERM, reset_exit);
-    signal(SIGILL, error_dump);
-    signal(SIGIOT, error_dump);
-    signal(SIGFPE, error_dump);
+    mySignal(SIGHUP, reset_exit);
+    mySignal(SIGINT, reset_exit);
+    mySignal(SIGQUIT, reset_exit);
+    mySignal(SIGTERM, reset_exit);
+    mySignal(SIGILL, error_dump);
+    mySignal(SIGIOT, error_dump);
+    mySignal(SIGFPE, error_dump);
 #ifdef	SIGBUS
-    signal(SIGBUS, error_dump);
+    mySignal(SIGBUS, error_dump);
 #endif				/* SIGBUS */
-    /* signal(SIGSEGV, error_dump); */
+    /* mySignal(SIGSEGV, error_dump); */
 }
 
 
@@ -915,7 +871,7 @@ write1(char c)
     putc(c, ttyf);
 #endif				/* not JP_CHARSET */
 #ifdef SCREEN_DEBUG
-    fflush(ttyf);
+    flush_tty();
 #endif				/* SCREEN_DEBUG */
     return 0;
 }
@@ -1248,7 +1204,7 @@ int
 graph_ok(void)
 {
 #ifndef KANJI_SYMBOLS
-    if (no_graphic_char)
+    if (!UseGraphicChar)
 	return 0;
 #endif				/* not KANJI_SYMBOLS */
     return T_as[0] != 0 && T_ae[0] != 0 && T_ac[0] != 0;
@@ -1471,7 +1427,7 @@ refresh(void)
 #endif				/* JP_CHARSET */
     }
     MOVE(CurLine, CurColumn);
-    fflush(ttyf);
+    flush_tty();
 }
 
 void
@@ -1491,6 +1447,7 @@ clear(void)
     CurrentMode = C_ASCII;
 }
 
+#ifdef USE_RAW_SCROLL
 static void
 scroll_raw(void)
 {				/* raw scroll */
@@ -1582,6 +1539,7 @@ rscroll(int n)
 	}
     }
 }
+#endif
 
 #if 0
 void
@@ -1866,8 +1824,20 @@ term_title(char *s)
 {
     if (!fmInitialized)
         return;
-    if (title_str != NULL)
+    if (title_str != NULL) {
+#ifdef __CYGWIN__
+	if (isLocalConsole && title_str == CYGWIN_TITLE) {
+	    Str buff;
+	    buff = Sprintf(title_str, s);
+	    if (buff->length > 1024) {
+		Strtruncate(buff, 1024);
+	    }
+	    SetConsoleTitle(buff->ptr);
+	}
+	else if (isLocalConsole || !isWinConsole)
+#endif
         fprintf(ttyf, title_str, s);
+    }
 }
 
 char
@@ -1876,7 +1846,7 @@ getch(void)
     char c;
 
     while (
-#ifdef __CYGWIN__
+#ifdef SUPPORT_WIN9X_CONSOLE_MBCS
 	      read_win32_console(&c, 1)
 #else
 	      read(tty, &c, 1)
@@ -1894,7 +1864,7 @@ getch(void)
 #ifdef USE_MOUSE
 #ifdef USE_GPM
 char
-wgetch(void)
+wgetch(void *p)
 {
     char c;
 
@@ -1997,12 +1967,12 @@ skip_escseq(void)
     }
 }
 
-void
+int
 sleep_till_anykey(int sec, int purge)
 {
     fd_set rfd;
     struct timeval tim;
-    int er, c;
+    int er, c, ret;
     TerminalMode ioval;
 
     TerminalGet(tty, &ioval);
@@ -2014,13 +1984,8 @@ sleep_till_anykey(int sec, int purge)
     FD_ZERO(&rfd);
     FD_SET(tty, &rfd);
 
-    if (
-#ifdef __CYGWIN__
-	   select_win32_console(tty + 1, &rfd, 0, 0, &tim)
-#else
-	   select(tty + 1, &rfd, 0, 0, &tim)
-#endif
-	   > 0 && purge) {
+    ret = select(tty + 1, &rfd, 0, 0, &tim);
+    if (ret > 0 && purge) {
 	c = getch();
 	if (c == ESC_CODE)
 	    skip_escseq();
@@ -2030,12 +1995,15 @@ sleep_till_anykey(int sec, int purge)
 	printf("Error occured: errno=%d\n", errno);
 	reset_exit(SIGNAL_ARGLIST);
     }
+    return ret;
 }
 
 #ifdef USE_MOUSE
 
 #define XTERM_ON   {fputs("\033[?1001s\033[?1000h",ttyf); flush_tty();}
 #define XTERM_OFF  {fputs("\033[?1000l\033[?1001r",ttyf); flush_tty();}
+#define CYGWIN_ON  {fputs("\033[?1000h",ttyf); flush_tty();}
+#define CYGWIN_OFF {fputs("\033[?1000l",ttyf); flush_tty();}
 
 #ifdef USE_GPM
 /* Linux console with GPM support */
@@ -2117,13 +2085,13 @@ mouse_init()
 	}
 #endif				/* defined(FBIO_MODEINFO) ||
 				 * defined(CONS_MODEINFO) */
-	signal(SIGUSR2, SIG_IGN);
+	mySignal(SIGUSR2, SIG_IGN);
 	mi.operation = MOUSE_MODE;
 	mi.u.mode.mode = 0;
 	mi.u.mode.signal = SIGUSR2;
 	sysm_handler = NULL;
 	if (ioctl(tty, CONS_MOUSECTL, &mi) != -1) {
-	    signal(SIGUSR2, sysmouse);
+	    mySignal(SIGUSR2, sysmouse);
 	    mi.operation = MOUSE_SHOW;
 	    ioctl(tty, CONS_MOUSECTL, &mi);
 	    sysm_handler = sysm_process_mouse;
@@ -2161,6 +2129,11 @@ mouse_init()
     if (is_xterm & NEED_XTERM_ON) {
 	XTERM_ON;
     }
+#ifdef __CYGWIN__
+    else if (is_xterm & NEED_CYGWIN_ON) {
+	CYGWIN_ON;
+    }
+#endif
     mouseActive = 1;
 }
 
@@ -2172,6 +2145,11 @@ mouse_end()
     if (is_xterm & NEED_XTERM_OFF) {
 	XTERM_OFF;
     }
+#ifdef __CYGWIN__
+    else if (is_xterm & NEED_CYGWIN_OFF) {
+	CYGWIN_OFF;
+    }
+#endif
     mouseActive = 0;
 }
 
@@ -2197,7 +2175,8 @@ mouse_inactive()
 void
 flush_tty()
 {
-    fflush(ttyf);
+    if (ttyf)
+	fflush(ttyf);
 }
 
 #ifdef USE_IMAGE
